#!/usr/bin/env python

"""Run the external MAP model.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as dsp

import thorns as th

import cochlea
from cochlea.external import run_matlab_auditory_periphery

def main():

    fs = 48e3

    ### Make sound
    tmax = 0.03
    t = np.arange(0, tmax, 1/fs)
    s = dsp.chirp(t, 80, t[-1], 16000)
    sound = cochlea.set_dbspl(s, 50)



    ### Run model
    anf = run_matlab_auditory_periphery(
        sound,
        fs,
        anf_num=(100,50,20),
        cf=(125, 16000, 80),
        seed=0,
    )



    ### Accumulate spike trains
    anf_acc = th.accumulate(anf, keep=['cf', 'duration'])
    anf_acc.sort('cf', ascending=False, inplace=True)



    ### Plot auditory nerve response
    fig, ax = plt.subplots(2, 1, sharex=True)
    th.plot_signal(
        signal=sound,
        fs=fs,
        ax=ax[0]
    )
    th.plot_neurogram(
        anf_acc,
        fs,
        ax=ax[1]
    )
    plt.show()



if __name__ == "__main__":
    main()



from __future__ import division, print_function, absolute_import

import numpy as np
import scipy.signal as dsp

from cochlea.holmberg2007.bm_pars import (
    real_freq_map,
    S_ST,
    S_ED,
    C_eardrum,
    outer_ear_a_48kHz,
    outer_ear_b_48kHz,
    outer_ear_a_100kHz,
    outer_ear_b_100kHz,
    delay_time,
)
from . _traveling_waves import (
    run_bm_wave,
    run_lcr4,
    run_ihcrp,
    run_ihc_meddis2000,
    run_an_sg_carney_holmberg2007
)


# Input signal should be multiplied by this factor
scaling_factor = S_ST * S_ED



def run_middle_ear_filter_orig(signal, fs):
    """Middle ear filter designed using digital wave techinique."""
    assert fs == 48000

    R2_ME = 1. / (2. * fs * C_eardrum)
    R1 = 1. / (2. * np.pi * C_eardrum * 1e3)

    g1_ME = (R2_ME - R1) / (R2_ME + R1)
    Z_ME=0

    out = np.zeros(len(signal))

    for i,samp in enumerate(signal):
        b2 = samp + g1_ME * (samp - Z_ME)
        Z_ME_b = Z_ME
        Z_ME = b2
        out[i] = ((Z_ME_b - b2) / R2_ME)

    return out


def _calc_middle_ear_coefs(fs):
    # Digital wave filter coefficients
    R2_ME = 1. / (2. * fs * C_eardrum)
    R1 = 1. / (2. * np.pi * C_eardrum * 1e3)
    g1_ME = (R2_ME - R1) / (R2_ME + R1)
    Z_ME=0

    # Standard filter coefficients
    b = [(-1-g1_ME), (1+g1_ME)]
    a = [R2_ME, R2_ME*g1_ME]

    return b, a


def run_middle_ear_filter(signal, fs):
    """Run middle ear filter model."""

    b,a = _calc_middle_ear_coefs(fs)

    return dsp.lfilter(b, a, signal)


def _calc_outer_ear_coefs(fs):
    if fs == 48000:
        a = outer_ear_a_48kHz
        b = outer_ear_b_48kHz
    elif fs == 100000:
        a = outer_ear_a_100kHz
        b = outer_ear_b_100kHz
    else:
        assert False, "Invalid sampling frequency: fs"

    return b, a


def run_outer_ear_filter(signal, fs):
    """Run outer ear filter."""

    b, a = _calc_outer_ear_coefs(fs)

    return dsp.lfilter(b, a, signal)



def get_nearest_cf(cf):
    """Return nearest valid frequency relative to `cf`."""

    idx = np.argmin(np.abs(real_freq_map - cf))
    nearest_cf = real_freq_map[idx]

    return nearest_cf



"""Cochlear filter tuning.

"""

from __future__ import division, print_function, absolute_import

import numpy as np
import pandas as pd

import thorns as th

from . threshold_rate import calc_spont_threshold, calc_threshold


def calc_tuning(
        model,
        cf,
        freqs=None,
        model_pars=None,
        map_backend=None,
):
    """Calculate runing of the cochlea at `cf`."""

    if freqs is None:
        freqs = np.logspace(np.log10(cf/2), np.log10(cf*2), 32)


    spont_rate = th.util.cache(calc_spont_threshold)(
        model=model,
        cf=cf,
        model_pars=model_pars
    )


    space = {
        'freq': freqs
    }

    kwargs = {
        'model': model,
        'model_pars': model_pars,
        'spont_rate': spont_rate,
        'cf': cf,
    }



    thresholds = th.util.map(
        calc_threshold,
        space,
        kwargs=kwargs,
        backend=map_backend,
    )

    thresholds.columns = ['threshold']

    return thresholds



from __future__ import division, print_function, absolute_import

import itertools
import numpy as np
import pandas as pd

from . import _pycat


def run_zilany2009(
        sound,
        fs,
        anf_num,
        cf,
        seed,
        cohc=1,
        cihc=1,
        powerlaw='approximate',
):
    """Run the inner ear model by [Zilany2009]_.

    This model is based on the original implementation provided by the
    authors.  The MEX specific code was replaced by Python code in C
    files.  We also compared the outputs of both implementation (see
    the `tests` directory for unit tests).


    Parameters
    ----------
    sound : array_like
        Input sound signal.
    fs : float
        Sampling frequency of the signal.
    anf_num : tuple
        The desired number of auditory nerve fibers per frequency
        channel (CF), (HSR#, MSR#, LSR#).  For example, (100, 75, 25)
        means that we want 100 HSR fibers, 75 MSR fibers and 25 LSR
        fibers per CF.
    cf : float or array_like or tuple
        The center frequency(s) of the simulated auditory nerve fibers.
        If float, then defines a single frequency channel.  If
        array_like (e.g. list or ndarray), then the frequencies are
        used.  If tuple, then must have exactly 3 elements (min_cf,
        max_cf, num_cf) and the frequencies are calculated using the
        Greenwood function.
    seed : int
        Random seed.
    cohc : flaot, optional
        Outer hair cell impairment parameter, must be between <0; 1>.
    cihc : flaot, optional
        Inner hair cell impairment parameter, must be between <0; 1>.
    powerlaw: {'approximate', 'actual'}, optional
        Type of the power-law implementation.


    Returns
    -------
    spike_trains
        Auditory nerve spike trains.


    Notes
    -----
    The fractorial Gausian noise from the oryginal implementation is
    disabled at the moment.



    References
    ----------
    If you are using results of this or modified version of the model
    in your research, please cite [Zilany2009]_.


    .. [Zilany2009] Zilany, M. S., Bruce, I. C., Nelson, P. C., &
       Carney, L. H. (2009). A phenomenological model of the synapse
       between the inner hair cell and auditory nerve: long-term
       adaptation with power-law dynamics. The Journal of the
       Acoustical Society of America, 126(5), 2390-2412.

    """
    assert np.max(sound) < 1000, "Signal should be given in Pa"
    assert sound.ndim == 1


    np.random.seed(seed)


    cfs = _calc_cfs(cf)


    ### Run Middle Ear filter
    meout = _pycat.run_middle_ear_filter(
        signal=sound,
        fs=fs
    )


    channel_args = [
        {
            'signal': meout,
            'cf': freq,
            'fs': fs,
            'cohc': cohc,
            'cihc': cihc,
            'anf_num': anf_num,
            'powerlaw': powerlaw,
        }
        for freq in cfs
    ]


    ### Run model for each channel
    nested = map(
        _run_channel,
        channel_args
    )


    ### Unpack the results
    trains = itertools.chain(*nested)
    spike_trains = pd.DataFrame(list(trains))


    if isinstance(np.fft.fftpack._fft_cache, dict):
        np.fft.fftpack._fft_cache = {}

    return spike_trains




def _run_channel(args):

    fs = args['fs']
    cf = args['cf']
    signal = args['signal']
    cohc = args['cohc']
    cihc = args['cihc']
    powerlaw = args['powerlaw']
    anf_num = args['anf_num']


    ### Run BM, IHC
    vihc = _pycat.run_ihc(
        signal=signal,
        cf=cf,
        fs=fs,
        cohc=float(cohc),
        cihc=float(cihc)
    )


    duration = len(vihc) / fs
    anf_types = np.repeat(['hsr', 'msr', 'lsr'], anf_num)

    synout = {}
    trains = []
    for anf_type in anf_types:

        if (anf_type not in synout):
            ### Run synapse
            synout[anf_type] = _pycat.run_synapse(
                fs=fs,
                vihc=vihc,
                cf=cf,
                anf_type=anf_type,
                powerlaw=powerlaw,
                ffGn=False
            )

        ### Run spike generator
        spikes = _pycat.run_spike_generator(
            synout=synout[anf_type],
            fs=fs,
        )


        trains.append({
            'spikes': spikes,
            'duration': duration,
            'cf': args['cf'],
            'type': anf_type
        })


    return trains





def _calc_cfs(cf):

    if np.isscalar(cf):
        cfs = [float(cf)]

    elif isinstance(cf, tuple):
        # Based on GenerateGreenwood_CFList() from DSAM
        # Liberman (1982)
        aA = 456
        k = 0.8
        a = 2.1

        freq_min, freq_max, freq_num = cf

        xmin = np.log10( freq_min / aA + k) / a
        xmax = np.log10( freq_max / aA + k) / a

        x_map = np.linspace(xmin, xmax, freq_num)
        cfs = aA * ( 10**( a*x_map ) - k)

    elif isinstance(cf, list) or isinstance(cf, np.ndarray):
        cfs = cf

    else:
        raise RuntimeError("CF must be a scalar, a tuple or a list.")

    return cfs



"""This implementation is based on the oryginal MATLAB implementation.

TODO: Unit test against the oryginal.

"""

from __future__ import division, print_function, absolute_import

import numpy as np
from numpy.random import randn
from scipy.signal import resample
from numpy.fft import fft, ifft

# sigma has been removed as it is never read
def ffGn(N, tdres, Hinput, mu):


    assert (N > 0)
    assert (tdres < 1)
    assert (Hinput >= 0) and (Hinput <= 2)


    # Downsampling No. of points to match with those of Scott jackson (tau 1e-1)
    resamp = int(np.ceil(1e-1 / tdres))
    nop = N
    N = int(np.ceil(N / resamp) + 1)
    if N < 10:
        N = 10

    # Determine whether fGn or fBn should be produced.
    if Hinput <= 1:
        H = Hinput
        fBn = 0
    else:
        H = Hinput - 1
        fBn = 1



    # Calculate the fGn.
    if H == 0.5:
        # If H=0.5, then fGn is equivalent to white Gaussian noise.
        y = randn(N)
    else:
        # TODO: make variables persistant
        Nfft = int(2 ** np.ceil(np.log2(2*(N-1))))
        NfftHalf = np.round(Nfft / 2)

        k = np.concatenate( (np.arange(0,NfftHalf), np.arange(NfftHalf,0,-1)) )
        Zmag = 0.5 * ( (k+1)**(2*H) -2*k**(2*H) + np.abs(k-1)**(2*H) )

        Zmag = np.real(fft(Zmag))
        assert np.all(Zmag >= 0)

        Zmag = np.sqrt(Zmag)

        Z = Zmag * (randn(Nfft) + 1j*randn(Nfft))

        y = np.real(ifft(Z)) * np.sqrt(Nfft)

        y = y[0:N]

        # Convert the fGn to fBn, if necessary.
        if fBn == 1:
            y = np.cumsum(y)


        # Resampling to match with the AN model
        y = resample(y, resamp*len(y))


        if mu < 0.5:
            sigma = 5
        elif mu < 18:
            sigma = 50          # 7 when added after powerlaw
        else:
            sigma = 200         # 40 when added after powerlaw


        y = y*sigma

        return y[0:nop]


def main():
    y = ffGn(10, 1e-1, 0.2, 1, debug=True)
    print(y)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo shows show to calculate synchronization of an auditory
model at different characteristic frequencies.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea
from cochlea.stats import calc_synchronization


def main():

    # vss = calc_synchronization(
    #     model=cochlea.run_holmberg2007,
    #     cfs=cochlea.holmberg2007.real_freq_map[10::5],
    #     model_pars={'fs': 48e3}
    # )

    vss = calc_synchronization(
        model=cochlea.run_zilany2014,
        model_pars={'species': 'human'}
    )

    print(vss)

    vss.plot(logx=True)

    plt.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

import os
from os.path import join

import scipy.io

import cochlea.zilany2009._pycat as pycat


DATADIR = os.path.join(
    os.path.dirname(__file__),
    'data_zilany2009'
)



def test_ihc():

    m = scipy.io.loadmat(
        join(DATADIR, 'zilany2009.mat'),
        squeeze_me=True
    )
    fs = float(m['Fs'])
    cf = float(m['CF'])
    sound = m['sound'].astype(float)
    vihc_target = m['vihc']

    meout = pycat.run_middle_ear_filter(
        signal=sound,
        fs=fs
    )
    vihc = pycat.run_ihc(
        signal=meout,
        cf=cf,
        fs=fs,
        cohc=1.,
        cihc=1.
    )

    assert_almost_equal(
        vihc,
        vihc_target,
        decimal=15
    )


def test_synapse():
    """test_synapse()

    This function has problems, because it's using matlab
    implementation of `resample' and it's difficult to exactly
    reproduce the output.

    During the generation of mat-file the `ffGn' was replaced by
    `zeros' and Matlab's `resample' by self-made downsampling function
    equivalent to python's implementation:

    function resampled = resample_fake(X, P, Q)
        b = fir1(Q, 1/Q);
        a = [1];
        filtered = filtfilt(b, a, X);
        resampled = filtered(1:Q:end);

    """
    m = scipy.io.loadmat(
        join(DATADIR, 'zilany2009.mat'),
        squeeze_me=True
    )
    fs = float(m['Fs'])
    cf = float(m['CF'])
    vihc = m['vihc']
    synout_target = m['synout']

    synout = pycat.run_synapse(
        fs=fs,
        vihc=vihc,
        cf=cf,
        anf_type='hsr',
        powerlaw='approximate',
        ffGn=False
    )

    assert_almost_equal(
        synout,
        synout_target,
        decimal=9
    )



#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

import pytest

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
import cochlea


def test_invertable():
    """Test if the greenwood function is invertable.

    """
    freq = np.linspace(20, 20e3, 100)
    x = cochlea.greenwood_inverse(freq, 'human')
    freq_greenwood = cochlea.greenwood(x, 'human')

    assert_almost_equal(freq, freq_greenwood)


def test_function():
    """Compare function results with correct value.

    """
    freq_0 = cochlea.greenwood(0, 'human')

    freq_0_target = 165.4 * (1 - 0.88)

    assert_equal(freq_0, freq_0_target)


def test_parameters():
    """Test if providing parameters results in the same as providing species.

    """
    x = np.linspace(0, 35e-3, 100)

    freq = cochlea.greenwood(x, A=165.4, a=60, k=0.88)
    freq_human = cochlea.greenwood(x, 'human')

    x = cochlea.greenwood_inverse(freq, A=165.4, a=60, k=0.88)
    x_human = cochlea.greenwood_inverse(freq, 'human')

    assert_equal(freq, freq_human)
    assert_equal(x, x_human)


def test_too_long():
    with pytest.raises(ValueError):
        cochlea.greenwood(40e-3, 'human')



# -*- coding: utf-8 -*-

from __future__ import division, print_function, absolute_import
from __future__ import unicode_literals

import pytest

from numpy.testing import assert_equal
import pandas as pd

import cochlea

@pytest.mark.skipif('True')
def test_make_brian_group():

    trains = pd.DataFrame([
        {'duration': 100e-3, 'spikes': [10e-3, 30e-3], 'cf': 333},
        {'duration': 100e-3, 'spikes': [40e-3, 50e-3], 'cf': 333},
        {'duration': 100e-3, 'spikes': [60e-3, 70e-3], 'cf': 333},
    ])


    group = cochlea.make_brian_group(trains)

    assert_equal(len(trains), len(group))



from __future__ import division, print_function, absolute_import
from __future__ import unicode_literals

import numpy as np
import pandas as pd
import itertools

from . import traveling_waves as tw


def run_holmberg2007_vesicles(
        sound,
        fs,
        anf_num,
        seed,
        cf=None,
):
    """Run the inner ear model by [Holmberg2007]_ in the quantal mode and
    return vesicles instead of spikes.

    Parameters
    ----------
    sound : array_like
        Input sound signal.
    fs : float
        Sampling frequency of the sound.
    anf_num : tuple
        Number of auditory nerve fibers per channel (HSR#, MSR#, LSR#).
    seed : int
        Random seed.
    cf : float or array_like or None, optional
        Characteristic frequencies.  If `None`, then calculate all 100
        predefined CFs.  CFs must be a subset of `real_freq_map`.


    Returns
    -------
    pd.DataFrame with vesicle timings
        Vesicles in the format of spike_trains.


    References
    ----------
    If you are using results of this or modified version of the model
    in your research, please cite [Holmberg2007]_.

    .. [Holmberg2007] Holmberg, M. (2007). Speech Encoding in the
       Human Auditory Periphery: Modeling and Quantitative Assessment
       by Means of Automatic Speech Recognition. PhD thesis, Technical
       University Darmstadt.

    """

    assert np.max(np.abs(sound)) < 1000, "Signal should be given in Pa"
    assert sound.ndim == 1
    assert fs == 48e3

    np.random.seed(seed)

    if cf is None:
        cfs = tw.real_freq_map
    elif np.isscalar(cf):
        cfs = [cf]
    else:
        cfs = cf


    assert set(cfs) <= set(tw.real_freq_map), set(cfs) - set(tw.real_freq_map)


    duration = len(sound) / fs



    ### Outer ear filter
    sound_oe = tw.run_outer_ear_filter(sound, fs)

    ### Middle ear filter
    sound_me = tw.run_middle_ear_filter(sound_oe, fs)

    ### Scaling
    sound_scaled = sound_me * tw.scaling_factor

    ### Basilar membrane
    xbm = tw.run_bm_wave(sound_scaled, fs)


    ihcrp = {}
    for cf in cfs:
        ### Amplification
        lcr4 = tw.run_lcr4(xbm[cf], fs, cf)

        ### Delay correction (1/cf)
        sec = np.where(tw.real_freq_map == cf)[0][0]
        lcr4_rolled = np.roll(
            lcr4,
            -int(np.round(tw.delay_time[99-sec]*fs))
        )

        ### IHCRP
        ihcrp[cf] = tw.run_ihcrp(lcr4_rolled, fs, cf)



    anf_types = np.repeat(['hsr', 'msr', 'lsr'], anf_num)


    ihc_meddis2000_pars = {
        'hsr': {
            'gamma_ca': 130,
            'beta_ca': 400,
            'tau_m': 1e-4,
            'gmax_ca': 4.5e-9,
            'e_ca': 0.066,
            'tau_ca': 1e-4,
            'perm_ca0': 0,
            'perm_z': 2e32,
            'power_ca': 3,
            'replenish_rate_y': 10,
            'loss_rate_l': 2580,
            'recovery_rate_r': 6580,
            'reprocess_rate_x': 66.3,
            'max_free_pool': 8,
        },
        'msr': {
            'gamma_ca': 130,
            'beta_ca': 400,
            'tau_m': 1e-4,
            'gmax_ca': 4.25e-9,
            'e_ca': 0.066,
            'tau_ca': 1e-4,
            'perm_ca0': 2.5e-11,
            'perm_z': 2e32,
            'power_ca': 3,
            'replenish_rate_y': 10,
            'loss_rate_l': 2580,
            'recovery_rate_r': 6580,
            'reprocess_rate_x': 66.3,
            'max_free_pool': 9,
        },
        'lsr': {
            'gamma_ca': 130,
            'beta_ca': 400,
            'tau_m': 1e-4,
            'gmax_ca': 2.75e-9,
            'e_ca': 0.066,
            'tau_ca': 1e-4,
            'perm_ca0': 4.2e-11,
            'perm_z': 2e32,
            'power_ca': 3,
            'replenish_rate_y': 10,
            'loss_rate_l': 2580,
            'recovery_rate_r': 6580,
            'reprocess_rate_x': 66.3,
            'max_free_pool': 6,
        }
    }

    psps = {}
    trains = []
    for cf,anf_type in itertools.product(ihcrp.keys(),anf_types):

        ### IHC and Synapse
        psp = tw.run_ihc_meddis2000(
            ihcrp=ihcrp[cf],
            fs=fs,
            syn_mode='quantal',
            **ihc_meddis2000_pars[anf_type]
        )


        psp = psp.astype(int)
        t = np.arange(0, len(psp)) / fs
        vesicles = np.repeat(t, psp)


        trains.append({
            'vesicles': vesicles,
            'duration': duration,
            'cf': cf,
            'type': anf_type
        })


    vesicle_trains = pd.DataFrame(trains)

    return vesicle_trains



#!/usr/bin/env python
"""Run inner ear model from [Zilany2014]_.

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as dsp

import thorns as th
import cochlea

def main():

    fs = 100e3

    # Make sound
    t = np.arange(0, 0.1, 1/fs)
    s = dsp.chirp(t, 80, t[-1], 20000)
    s = cochlea.set_dbspl(s, 50)
    pad = np.zeros(int(10e-3 * fs))
    sound = np.concatenate( (s, pad) )

    # Run model
    anf = cochlea.run_zilany2014(
        sound,
        fs,
        anf_num=(100,0,0),
        cf=(125, 20000, 100),
        seed=0,
        powerlaw='approximate',
        species='human',
    )

    # Accumulate spike trains
    anf_acc = th.accumulate(anf, keep=['cf', 'duration'])
    anf_acc.sort_values('cf', ascending=False, inplace=True)

    # Plot auditory nerve response
    fig, ax = plt.subplots(2,1)
    th.plot_signal(
        signal=sound,
        fs=fs,
        ax=ax[0]
    )
    th.plot_neurogram(
        anf_acc,
        fs,
        ax=ax[1]
    )
    plt.show()



if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo calculates tuning of cochlear filters.

On a fast computer it will take around 2 minutes to finish.  If you
set shell variable THmap=m, then it will use multiple cores.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea
from cochlea.stats import calc_tuning


def main():

    ths_human = calc_tuning(
        model=cochlea.run_zilany2014,
        cf=5e3,
        model_pars={'species': 'human'}
    )

    print(ths_human)


    fig,ax = plt.subplots()

    ths_human.plot(logx=True, ax=ax)

    ax.set_title("Zilany et al. (2014)")
    ax.legend(loc='best')

    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division
from __future__ import print_function


raise NotImplementedError


import numpy as np
import pandas as pd
import logging

import thorns as th
import thorns.waves as wv



def calc_spont_threshold(model, model_pars=None):

    if model_pars is None:
        model_pars = {}

    pars = dict(model_pars)

    fs = pars.setdefault('fs', 100e3)
    cf = pars.setdefault('cf', 1e3)

    tmax = 250e-3
    s = np.zeros(fs*tmax)

    anf = model(
        sound=s,
        anf_num=(10000,0,0),
        seed=0,
        **pars
    )

    sis = [th.calc_si(train,cf) for _,train in anf.iterrows()]
    sis = np.array(sis)

    threshold = sis.mean() + sis.std()

    return threshold



def error_func(dbspl, model, cf, model_pars, spont_si):

    if model_pars is None:
        model_pars = {}

    pars = dict(model_pars)

    fs = pars.setdefault('fs', 100e3)

    tone_duration = 250e-3
    onset = 15e-3

    s = wv.make_ramped_tone(
        fs,
        cf,
        duration=tone_duration,
        ramp=2.5e-3,
        pad=0,
        dbspl=dbspl
    )

    anf = model(
        sound=s,
        anf_num=(1000, 0, 0),
        cf=cf,
        seed=0,
        **pars
    )

    trains = th.trim(anf, onset, None)
    si = th.calc_si(trains, cf)

    error = si - spont_si

    logging.debug("{} {} {}".format(dbspl, si, error))

    return error



def calc_hearing_threshold_si(model, cf, spont_si, model_pars=None):

    kwargs = {
        'model':model,
        'model_pars': model_pars,
        'cf': cf,
        'spont_si': spont_si
    }
    dbspl_opt = th.util.find_zero(
        error_func,
        kwargs=kwargs,
        x1=-100,
        x2=50,
        xtol=0.1
    )

    return dbspl_opt






def calc_hearing_thresholds_si(
        model,
        cfs=None,
        **model_pars
):

    if cfs is None:
        cfs = np.logspace(np.log10(100), np.log10(16000), 32)


    spont_si = th.util.apply(
        calc_spont_threshold,
        model=model,
        model_pars=model_pars
    )

    space = [
        {
            'model': model,
            'model_pars': model_pars,
            'spont_si': spont_si,
            'cf': cf
        }
        for cf in cfs
    ]


    thresholds = th.util.map(
        calc_hearing_threshold_si,
        space,
        # backend='multiprocessing'
    )
    thresholds = list(thresholds)

    thresholds = pd.Series(thresholds, index=cfs)

    return thresholds





def main():
    import cochlea
    import matplotlib.pyplot as plt

    spont_si = calc_spont_threshold(
        cochlea.run_zilany2009
    )
    print(spont_si)


    r = calc_hearing_threshold_si(
        model=cochlea.run_zilany2009,
        cf=4e3,
        spont_si=spont_si
    )

    print(r)
    exit()


    ths = calc_hearing_thresholds_si(
        model=cochlea.run_zilany2009
    )
    print(ths)

    ths.plot(logx=True)
    plt.show()



if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Run innear ear model by [Holmberg2007]_ in quantal mode and output
vesicle events instead of spikes.

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as dsp

import thorns as th

import cochlea

def main():

    fs = 48e3
    tmax = 0.1

    ### Make sound
    t = np.arange(0, tmax, 1/fs)
    s = np.zeros_like(t)


    ### Run model
    vesicle_trains = cochlea.run_holmberg2007_vesicles(
        s,
        fs,
        anf_num=(1,0,0),
        seed=0,
    )



    print(vesicle_trains)

    ### Need to rename a column: vesicles -> spikes, because that's
    ### the expected name by many functions in thorns
    trains = vesicle_trains.rename(columns={'vesicles': 'spikes'})

    ### Calculate average rate
    rate = th.firing_rate(trains)
    print()
    print("Spontanious rate:", rate)


    ### Raster plot
    th.plot_raster(trains)
    th.show()


if __name__ == "__main__":
    main()



"""Synchronization of inner ear models.

"""
from __future__ import division, absolute_import, print_function
from __future__ import unicode_literals

import numpy as np

import thorns as th
import thorns.waves as wv


def calc_synchronization(
        model,
        cfs=None,
        dbspls=None,
        model_pars=None,
        map_backend=None,
):
    """Calculate vector strength of an inner ear model.

    """
    if model_pars is None:
        model_pars = {}

    if cfs is None:
        cfs = np.logspace(np.log10(125), np.log10(16e3), 16)

    if dbspls is None:
        dbspls = [20, 40, 60]

    space = {
        'dbspl': dbspls,
        'cf': cfs,
    }

    kwargs = {
        'model': model,
        'model_pars': model_pars,
    }

    all_vss = th.util.map(
        _run_model,
        space,
        kwargs=kwargs,
        backend=map_backend,
    )

    vss = all_vss.reset_index()

    best_vss = vss.groupby('cf').max().drop('dbspl', axis=1)

    return best_vss


def _run_model(model, dbspl, cf, model_pars):

    duration = 100e-3
    onset = 10e-3

    fs = model_pars.setdefault('fs', 100e3)
    model_pars.setdefault('anf_num', (250, 250, 250))
    model_pars.setdefault('seed', 0)

    sound = wv.ramped_tone(
        fs=fs,
        freq=cf,
        duration=duration,
        pad=0,
        dbspl=dbspl
    )

    anf = model(
        sound=sound,
        cf=cf,
        **model_pars
    )

    # th.plot_raster(anf)
    # th.show()

    # We want to make sure the the output CF is equal to the desired
    # CF.
    real_cf, = np.unique(anf['cf'])
    assert real_cf == cf

    vss = {}
    for typ, group in anf.groupby('type'):
        trimmed = th.trim(group, onset, None)
        vs = th.vector_strength(trimmed, cf)
        vss[typ] = vs

    return vss



"""Rate-level charactersitics for inner ear models.

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import pandas as pd

import thorns as th
import thorns.waves as wv


def calc_rate_level(
        model,
        dbspls=None,
        cf=1000,
        model_pars=None,
        tone_duration=100e-3,
):
    """Calculate rate-level characteristic of an auditory model.

    Parameters
    ----------
    dbspls : array_like, optional
        An array of sound levels (dB SPL) for which to calculate
        rates.

    """
    if model_pars is None:
        model_pars = {}

    if dbspls is None:
        dbspls = np.arange(-10, 100, 5)

    space = {
        'dbspl': dbspls
    }

    kwargs = {
        'model': model,
        'cf': cf,
        'tone_duration': tone_duration,
        'model_pars': model_pars,
    }

    rates = th.util.map(
        _run_model,
        space,
        kwargs=kwargs,
    )

    return rates



def _run_model(model, dbspl, cf, model_pars, tone_duration):

    onset = 10e-3
    assert tone_duration > onset

    fs = model_pars.setdefault('fs', 100e3)
    model_pars.setdefault('anf_num', (250,250,250))
    model_pars.setdefault('seed', 0)

    sound = wv.ramped_tone(
        fs=fs,
        freq=cf,
        duration=tone_duration,
        pad=0,
        dbspl=dbspl
    )

    anf = model(
        sound=sound,
        cf=cf,
        **model_pars
    )

    rates = {}
    for typ,trains in anf.groupby('type'):
        trimmed = th.trim(trains, onset, None)
        rate = th.firing_rate(trimmed)
        rates[typ] = rate

    return rates



#!/usr/bin/env python

"""Run innear ear model by [Holmberg2007]_.

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as dsp

import thorns as th

import cochlea

def main():

    fs = 48e3

    ### Make sound
    t = np.arange(0, 0.1, 1/fs)
    s = dsp.chirp(t, 80, t[-1], 20000)
    s = cochlea.set_dbspl(s, 50)
    s = np.concatenate( (s, np.zeros(int(10e-3 * fs))) )



    ### Run model
    anf_trains = cochlea.run_holmberg2007(
        s,
        fs,
        anf_num=(100,0,0),
        seed=0,
    )


    ### Plot auditory nerve response
    anf_acc = th.accumulate(anf_trains, keep=['cf', 'duration'])
    anf_acc.sort('cf', ascending=False, inplace=True)


    fig, ax = plt.subplots(2,1)
    th.plot_signal(
        signal=s,
        fs=fs,
        ax=ax[0]
    )
    th.plot_neurogram(
        anf_acc,
        fs,
        ax=ax[1]
    )
    plt.show()



if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo calculates rate based hearing threshold of an innear ear
model.

On a fast computer it will take around 1 min to finish.  If you set
shell variable THmap=m, then it will use multiple cores.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea
from cochlea.stats import calc_thresholds_rate


def main():

    ths = calc_thresholds_rate(
        model=cochlea.run_zilany2014,
        model_pars={'species': 'cat'}
    )
    print(ths)

    ths.plot(logx=True)
    plt.show()


if __name__ == "__main__":
    main()



# Copyright 2009-2014 Marek Rudnicki
#
# This file is part of cochlea.
#
# cochlea is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cochlea is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cochlea.  If not, see <http://www.gnu.org/licenses/>.

"""Innear ear models in Python.

"""

from __future__ import division, print_function, absolute_import


import warnings
import sys
import numpy as np

from cochlea.zilany2009 import run_zilany2009
from cochlea.holmberg2007 import run_holmberg2007
from cochlea.holmberg2007 import run_holmberg2007_vesicles
from cochlea.zilany2014 import run_zilany2014
from cochlea.zilany2014 import run_zilany2014_rate


from cochlea.holmberg2007 import real_freq_map as freq_map_holmberg2007
from cochlea.holmberg2007 import get_nearest_cf as get_nearest_cf_holmberg2007


__version__ = "1.4"


# Check if running 64-bit version of Python
if sys.maxsize <= 2**32:
    warnings.warn("cochlea: it seems that you are using 32-bit" +
                  "version of Python." +
                  "If you experience issues, please switch to 64-bit version.")


def set_dbspl(signal, dbspl):
    """Rescale the signal to a new level in dB SPL.

    Parameters
    ----------
    signal : array_like
        Signal for scaling.
    dbspl : float
        Desired level of the output signal in dB SPL.

    Returns
    -------
    array_like
        Scaled version of the original signal.

    """
    p0 = 20e-6
    rms = np.sqrt(np.sum(signal**2) / signal.size)

    scalled = signal * 10**(dbspl/20) * p0 / rms

    return scalled


def set_dba_isolet(signal, dba):
    p0 = 20e-6

    # value from miclib (precalculated for all ISOLET files)
    rms_dba = 0.02972401089

    scaled = signal * 10**(dba/20) * p0 / rms_dba

    return scaled


def make_brian_group(trains):
    """Create Brian's spike generator group from spike trains.

    Parameters
    ----------
    trains : spike_trains
        Input spike trains

    Returns
    -------
    brian.SpikeGeneratorGroup
        Brian's spike generator group.

    """
    import brian

    times = []
    indices = []
    for i, spikes in enumerate(trains['spikes']):
        times.append(spikes)
        indices.append(np.ones(len(spikes)) * i)

    indices = np.concatenate(indices)
    times = np.concatenate(times) * brian.second

    group = brian.SpikeGeneratorGroup(
        len(trains),
        spiketimes=(indices, times)
    )

    return group


_greenwood_pars  = {
    'human': {'A': 165.4, 'a': 60, 'k': 0.88, 'length': 35e-3},
    'cat': {'A': 456, 'a': 84, 'k': 0.8, 'length': 25e-3},
}


def greenwood(x, species=None, A=None, a=None, k=None):
    '''Greenwood function.

    Calculates the corresponding center frequency for a place on the
    basilar membrane.
    Center frequency and place on the basilar membrane can be
    connected by using the Greenwood equation [1].
    .. math:: cf = A(10^{x \cdot a) - k)
    where cf is the center frequency in Hz and x the place on the
    basilar membrane given from apex to base (apex = 0). The parameter
    A has the unit Hz, a has a unit of 1/m and k is unit less.

    Parameters
    ----------
    x : float or ndarray
        The position of the basilar membrane in meters. Apex = 0m
    species : str
        A string specifing the species ('human' or 'cat')
        coeffients from [1])
    A, x, k : float
        Specifying the parameters of the Greenwood equation
        instead of using the implemented species. If parameters
        are given, the species parameter is ignored.

    .. [1] Greenwood, D. D. (1990). A cochlear frequency-position
    function for several species--29 years later. The Journal of the
    Acoustical Society of America, 87(6)
    '''
    if species is not None:
        pars = _greenwood_pars[species]

        A = pars['A']
        a = pars['a']
        k = pars['k']

        if np.any(x > pars['length']):
            raise ValueError("Cochlea too long.")

    cf = A * (10**(x * a) - k)

    return cf


def greenwood_inverse(cf, species=None, A=None, a=None, k=None):
    '''Inverse Greenwood function

    Calculate the place on the basilar membrane from the corresponding
    center frequency.
    Center frequency and place on the basilar membrane can be
    connected by using the Greenwood equation [1].
    .. math:: x = \frac{log_{10}(cf / A + k)}{a}
    where cf is the center frequency in Hz and x the place on the
    basilar membrane given from apex to base (apex = 0). The parameter
    A has the unit Hz, a has a unit of 1/m and k is unit less.

    Parameters
    ----------
    cf : scalar or ndarray
        The center frequency in Hz
    species : str
        A string specifing the species ('human' or 'cat')
        coeffients from [1])
    A, x, k : float
        Specifying the parameters of the Greenwood equation
        instead of using the implemented species. If parameters
        are given, the species parameter is ignored.

    Returns
    -------
    scalar or ndarray
        The position on the basilar membrane in m where Apex = 0m

    .. [1] Greenwood, D. D. (1990). A cochlear frequency-position
    function for several species--29 years later. The Journal of the
    Acoustical Society of America, 87(6)
    '''
    if species is not None:
        pars = _greenwood_pars[species]

        A = pars['A']
        a = pars['a']
        k = pars['k']

    x = np.log10((cf / A) + k) / a

    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Run Zilany et al. (2014) model and plot a PSTH.

Unlike other examples, this script does not need any extra (unusual)
dependencies (e.g. thorns) for visualization.  It depends only on
matplotlib.  If you want to run more advanced examples, you will need
to install `thorns'.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea

def main():
    fs = 100e3
    cf = 8e3

    ### Make sound
    t = np.arange(0, 0.1, 1/fs)
    tone = np.sin(2*np.pi*t*cf)

    tone = cochlea.set_dbspl(tone, 20)

    pad = np.zeros(int(50e-3 * fs))
    sound = np.concatenate( (pad, tone, pad) )



    ### Run model
    anf = cochlea.run_zilany2014(
        sound,
        fs,
        anf_num=(200,0,0),
        cf=cf,
        seed=0,
        powerlaw='approximate',
        species='human'
    )


    print(anf.head(20))


    ### Plot PSTH
    all_spikes = np.concatenate(anf.spikes)
    tmax = anf.duration.max()

    bin_size = 1e-3

    fig,ax = plt.subplots()

    ax.hist(
        all_spikes,
        bins=int(tmax/bin_size),
        range=(0,tmax),
        weights=np.ones_like(all_spikes)/bin_size/len(anf)
    )

    ax.set_xlabel("Time [s]")
    ax.set_ylabel("Rate [spikes/s]")
    ax.set_title("PSTH")

    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

import os
from os.path import join

import cochlea.holmberg2007.traveling_waves as tw


DATADIR = os.path.join(
    os.path.dirname(__file__),
    'data_holmberg2007'
)


def test_bm_wave():

    data = np.load(join(DATADIR, 'holmberg2007.npz'))

    fs = data['fs']
    sound_scaled = data['sound_scaled']
    xbm_target = data['xbm']
    channel = int(data['channel'])
    cf = tw.real_freq_map[channel]

    xbm = tw.run_bm_wave(
        sound_scaled,
        fs
    )

    assert_equal(
        xbm[cf],
        xbm_target
    )





def test_lcr4():

    data = np.load(join(DATADIR,'holmberg2007.npz'))
    xbm = data['xbm']
    lcr4_target = data['lcr4']
    channel = int(data['channel'])
    cf = tw.real_freq_map[channel]

    lcr4 = tw.run_lcr4(
        xbm=xbm,
        fs=data['fs'],
        cf=cf
    )

    assert_equal(
        lcr4,
        lcr4_target
    )



def test_ihcrp():

    data = np.load(join(DATADIR,'holmberg2007.npz'))
    fs = data['fs']
    lcr4 = data['lcr4']
    ihcrp_target = data['ihcrp']
    channel = int(data['channel'])
    cf = tw.real_freq_map[channel]

    ihcrp = tw.run_ihcrp(
        xbm=lcr4,
        fs=fs,
        cf=cf
    )

    assert_almost_equal(
        ihcrp,
        ihcrp_target,
        decimal=16
    )




def test_ihc_meddis2000():
    """ test_ihc_meddis2000()
    Testing against the DSAM implementation

    """
    data = np.load(join(DATADIR, 'holmberg2007.npz'))

    fs = data['fs']
    ihcrp = data['ihcrp']
    psp_target = data['psp']

    psp = tw.run_ihc_meddis2000(
        ihcrp=ihcrp,
        fs=fs,
        gamma_ca=130,
        beta_ca=400,
        tau_m=1e-4,
        gmax_ca=4.5e-9,
        e_ca=0.066,
        tau_ca=1e-4,
        perm_ca0=0,
        perm_z=2e32,
        power_ca=3,
        replenish_rate_y=10,
        loss_rate_l=2580,
        recovery_rate_r=6580,
        reprocess_rate_x=66.3,
        max_free_pool=8,
    )

    assert_almost_equal(
        psp,
        psp_target,
        decimal=16
    )



# def test_an_sg_carney_holmberg2007():

#     data = np.load('data/holmberg2007.npz')

#     fs = data['fs']
#     psp = data['psp']

#     an = tw.run_an_sg_carney_holmberg2007(
#         psp=psp,
#         fs=fs,
#         c0=0.5,
#         c1=0.5,
#         s0=1e-3,
#         s1=12.5e-3,
#         refractory_period=0.75e-3
#     )

#     print(an)



"""Copyright 2009-2014 Marek Rudnicki

This file is part of cochlea.

cochlea is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

cochlea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with cochlea.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
from libc.stdlib cimport malloc
import scipy.signal as dsp
import ffGn_module

cimport numpy as np


cdef extern from "stdlib.h":
    void *memcpy(void *str1, void *str2, size_t n)


cdef extern from "catmodel.h":
    void IHCAN(
        double *px,
        double cf,
        int nrep,
        double tdres,
        int totalstim,
        double cohc,
        double cihc,
        double *ihcout
    )
    double Synapse(
        double *ihcout,
        double tdres,
        double cf,
        int totalstim,
        int nrep,
        double spont,
        double implnt,
        double sampFreq,
        double *synouttmp,
        int with_ffGn
    )
    int SpikeGenerator(
        double *synouttmp,
        double tdres,
        int totalstim,
        int nrep,
        double *sptime
    )


cdef extern from "Python.h":
    ctypedef int Py_intptr_t


cdef extern from "numpy/arrayobject.h":
    ctypedef Py_intptr_t npy_intp
    object PyArray_SimpleNewFromData(
        int nd,
        npy_intp* dims,
        int typenum,
        void* data
    )



np.import_array()




cdef public double* generate_random_numbers(long length):
    arr = np.random.rand(length)

    if not arr.flags['C_CONTIGUOUS']:
        arr = arr.copy(order='C')

    cdef double *data_ptr = <double *>np.PyArray_DATA(arr)
    cdef double *out_ptr = <double *>malloc(length * sizeof(double))
    memcpy(out_ptr, data_ptr, length*sizeof(double))

    return out_ptr




cdef public double* decimate(
    int k,
    double *signal,
    int q
):
    """Decimate a signal

    k: number of samples in signal
    signal: pointer to the signal
    q: decimation factor

    This implementation was inspired by scipy.signal.decimate.

    """
    # signal_arr will not own the data, signal's array has to be freed
    # after return from this function
    signal_arr = PyArray_SimpleNewFromData(
        1,                      # nd
        [k],                    # dims
        np.NPY_DOUBLE,          # typenum
        <void *>signal          # data
    )


    # resampled = dsp.resample(
    #     signal_arr,
    #     len(signal_arr) // q
    # )


    b = dsp.firwin(q+1, 1./q, window='hamming')
    a = [1.]

    filtered = dsp.filtfilt(
        b=b,
        a=a,
        x=signal_arr
    )

    resampled = filtered[::q]


    if not resampled.flags['C_CONTIGUOUS']:
        resampled = resampled.copy(order='C')


    # Copy data to output array
    cdef double *resampled_ptr = <double *>np.PyArray_DATA(resampled)
    cdef double *out_ptr = <double *>malloc(len(resampled)*sizeof(double))
    memcpy(out_ptr, resampled_ptr, len(resampled)*sizeof(double))

    return out_ptr


cdef public double* ffGn(int N, double tdres, double Hinput, double mu, bint with_ffGn):
    """ ffGn.py wrapper """

    if with_ffGn:
        a = ffGn_module.ffGn(N, tdres, Hinput, mu)
    else:
        a = np.zeros(N)

    if not a.flags['C_CONTIGUOUS']:
        a = a.copy(order='C')

    # Copy data to output array
    cdef double *ptr = <double *>np.PyArray_DATA(a)
    cdef double *out_ptr = <double *>malloc(len(a)*sizeof(double))
    memcpy(out_ptr, ptr, len(a)*sizeof(double))

    return out_ptr





def run_middle_ear_filter(
        np.ndarray[np.float64_t, ndim=1] signal,
        double fs
):

    # variables for middle-ear model
    megainmax=43;
    # double *mey1, *mey2, *mey3, meout,c1filterouttmp,c2filterouttmp,c1vihctmp,c2vihctmp;
    # double fp,C,m11,m12,m21,m22,m23,m24,m25,m26,m31,m32,m33,m34,m35,m36;
    tdres = 1 / fs

    fp = 1e3 # /* prewarping frequency 1 kHz */
    C = 2*np.pi*fp/np.tan(np.pi*fp*tdres)

    m11 = C/(C + 693.48)
    m12 = (693.48 - C)/C

    m21 = 1/(C**2 + 11053*C + 1.163e8)
    m22 = -2*C**2 + 2.326e8
    m23 = C**2 - 11053*C + 1.163e8
    m24 = C**2 + 1356.3*C + 7.4417e8
    m25 = -2*C**2 + 14.8834e8
    m26 = C**2 - 1356.3*C + 7.4417e8

    m31 = 1/(C**2 + 4620*C + 909059944)
    m32 = -2*C**2 + 2*909059944
    m33 = C**2 - 4620*C + 909059944
    m34 = 5.7585e5*C + 7.1665e7
    m35 = 14.333e7
    m36 = 7.1665e7 - 5.7585e5*C

    mey1 = dsp.lfilter([1,-1], [1/m11,m12], signal)
    mey2 = dsp.lfilter([m24,m25,m26], [1/m21,m22,m23], mey1)
    mey3 = dsp.lfilter([m34,m35,m36], [1/m31,m32,m33], mey2)

    meout = mey3 / megainmax

    return meout


def run_ihc(
        np.ndarray[np.float64_t, ndim=1] signal,
        double cf,
        double fs,
        double cohc=1.,
        double cihc=1.
):
    """Run BM / IHC model.

    signal: output of the middle ear filter [Pa]
    cf: characteristic frequency
    fs: sampling frequency
    cohc, cihc: degeneration parameters for IHC and OHC cells

    return: IHC receptor potential

    """
    assert (cf > 79.9) and (cf < 40e3), "Wrong CF: 80 <= cf < 40e3, CF = %s"%str(cf)
    assert (fs >= 100e3) and (fs <= 500e3), "Wrong Fs: 100e3 <= fs <= 500e3"
    assert (cohc >= 0) and (cohc <= 1), "0 <= cohc <= 1"
    assert (cihc >= 0) and (cihc <= 1), "0 <= cihc <= 1"


    # Input sound
    if not signal.flags['C_CONTIGUOUS']:
        signal = signal.copy(order='C')
    cdef double *signal_data = <double *>np.PyArray_DATA(signal)

    # Output IHC voltage
    ihcout = np.zeros( len(signal) )
    cdef double *ihcout_data = <double *>np.PyArray_DATA(ihcout)


    IHCAN(signal_data, cf, 1, 1.0/fs, len(signal), cohc, cihc, ihcout_data)


    return ihcout




def run_synapse(
        np.ndarray[np.float64_t, ndim=1] vihc,
        double cf,
        double fs,
        anf_type='hsr',
        powerlaw='actual',
        ffGn=True
):
    """Run synapse simulation.

    vihc: IHC receptor potential
    cf: characteristic frequency
    anf_type: auditory nerve fiber type ('hsr', 'msr' or 'lsr')
    powerlaw: implementation of the powerlaw ('actual', 'approximate')
    ffGn: enable/disable factorial Gauss noise generator

    return: PSTH from ANF

    """
    assert (cf > 79.9) and (cf < 40e3), "Wrong CF: 80 <= cf < 40e3, CF = %s"%str(cf)
    assert (fs >= 100e3) and (fs <= 500e3), "Wrong Fs: 100e3 <= fs <= 500e3"
    assert anf_type in ['hsr', 'msr', 'lsr'], "anf_type not hsr/msr/lsr"
    assert powerlaw in ['actual', 'approximate'], "powerlaw not actual/approximate"

    spont = {
        'hsr': 100.,
        'msr': 5.,
        'lsr': 0.1
    }

    powerlaw_map = {
        'actual': 1,
        'approximate': 0
    }


    # Input IHC voltage
    if not vihc.flags['C_CONTIGUOUS']:
        vihc = vihc.copy(order='C')
    cdef double *vihc_data = <double *>np.PyArray_DATA(vihc)


    # Output synapse data (spiking probabilities)
    synout = np.zeros_like(vihc)
    cdef double *synout_data = <double *>np.PyArray_DATA(synout)


    # Run synapse model
    Synapse(
        vihc_data,                   # px
        1.0/fs,                      # tdres
        cf,                          # cf
        len(vihc),                   # totalstim
        1,                           # nrep
        spont[anf_type],             # spont
        powerlaw_map[powerlaw],      # powerlaw
        10e3,                        # sampFreq
        synout_data,                 # synouttmp
        ffGn                         # ffGn
    )

    return synout





def run_spike_generator(
        np.ndarray[np.float64_t, ndim=1] synout,
        double fs
):
    """Run spike generator.

    synout: synapse output (sp/s)
    fs: sampling frequency

    return: sptime

    """
    # Input IHC voltage
    if not synout.flags['C_CONTIGUOUS']:
        synout = synout.copy(order='C')
    cdef double *synout_data = <double *>np.PyArray_DATA(synout)

    # Output spikes (signal)
    sptimes = np.zeros(int(np.ceil(len(synout)/0.00075/fs)))
    cdef double *sptimes_data = <double *>np.PyArray_DATA(sptimes)


    # Run synapse model
    SpikeGenerator(
        synout_data,            # synouttmp
        1./fs,                  # tdres
        len(synout),            # totalstim
        1,                      # nprep
        sptimes_data            # sptime
    )

    spikes = np.array(sptimes[sptimes != 0])

    return spikes





def sample_rate(fs, rates):

    rates = rates / fs
    events = np.zeros_like(rates)
    randoms = np.random.rand(*rates.shape)

    events[rates > randoms] = 1

    return events



from __future__ import division, print_function, absolute_import

import itertools
import numpy as np
import pandas as pd

from . import _zilany2014
from . util import calc_cfs

def run_zilany2014_rate(
        sound,
        fs,
        anf_types,
        cf,
        species,
        cohc=1,
        cihc=1,
        powerlaw='approximate',
        ffGn=False
):
    """Run the inner ear model by [Zilany2014]_.  Return mean firing rate
    of the auditory nerve fibers.


    Notes
    -----
    This implementation is was not used very much and may have some
    problems.  Use with caution!  (Like any implementation here, BTW)


    References
    ----------

    .. [Zilany2014] Zilany, M. S., Bruce, I. C., & Carney,
       L. H. (2014). Updated parameters and expanded simulation
       options for a model of the auditory periphery. The Journal of
       the Acoustical Society of America, 135(1), 283-286.

    """
    assert np.max(sound) < 1000, "Signal should be given in Pa"
    assert sound.ndim == 1
    assert species in ('cat', 'human')


    if isinstance(anf_types, str):
        anf_types = [anf_types]

    cfs = calc_cfs(cf, species)

    channel_args = [
        {
            'signal': sound,
            'cf': cf,
            'fs': fs,
            'cohc': cohc,
            'cihc': cihc,
            'anf_types': anf_types,
            'powerlaw': powerlaw,
            'species': species,
            'ffGn': ffGn,
        }
        for cf in cfs
    ]


    ### Run model for each channel
    nested_results = map(
        _run_channel,
        channel_args
    )

    results = list(itertools.chain(*nested_results))

    columns = pd.MultiIndex.from_tuples(
        [(r['anf_type'],r['cf']) for r in results],
        names=['anf_type','cf']
    )
    rates = np.array([r['rate'] for r in results]).T

    rates = pd.DataFrame(
        rates,
        columns=columns
    )

    if isinstance(np.fft.fftpack._fft_cache, dict):
        np.fft.fftpack._fft_cache = {}

    return rates




def _run_channel(args):

    fs = args['fs']
    cf = args['cf']
    signal = args['signal']
    cohc = args['cohc']
    cihc = args['cihc']
    powerlaw = args['powerlaw']
    anf_types = args['anf_types']
    species = args['species']
    ffGn = args['ffGn']


    ### Run BM, IHC
    vihc = _zilany2014.run_ihc(
        signal=signal,
        cf=cf,
        fs=fs,
        species=species,
        cohc=float(cohc),
        cihc=float(cihc)
    )


    duration = len(vihc) / fs


    rates = []
    for anf_type in anf_types:

        ### Run synapse
        synout = _zilany2014.run_synapse(
            fs=fs,
            vihc=vihc,
            cf=cf,
            anf_type=anf_type,
            powerlaw=powerlaw,
            ffGn=ffGn
        )

        rates.append({
            'rate': synout / (1 + 0.75e-3*synout),
            'cf': cf,
            'anf_type': anf_type
        })

    return rates



# Copyright L. H. Carney
# Copyright 1996-1997 L. P. O'Mard
# Copyright 2000-2001 Werner Hemmert
# Copyright 2002-2007 Marcus Holmberg
# Copyright 2009-2014 Marek Rudnicki

# This file is part of cochlea.

# cochlea is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# cochlea is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with cochlea.  If not, see <http://www.gnu.org/licenses/>.

"""This file is a Cython reimplementation of various C files from the
oryginal model [Holmberg2007]_

"""


from __future__ import division



import numpy as np
cimport numpy as np


cdef extern from "math.h":
    double sqrt(double x)
    double exp(double x)
    double fabs(double x)


import bm_pars




def run_bm_wave(
        np.ndarray[np.float64_t, ndim=1] signal,
        np.float64_t fs,
):

    assert fs == 48e3

    cdef np.ndarray[np.float64_t] Ls = bm_pars.Ls
    cdef np.ndarray[np.float64_t] Rs = bm_pars.Rs
    cdef np.ndarray[np.float64_t] Ct = bm_pars.Ct
    cdef np.ndarray[np.float64_t] Rbm = bm_pars.Rbm
    cdef np.ndarray[np.float64_t] Cbm = bm_pars.Cbm
    cdef np.ndarray[np.float64_t] Lbm = bm_pars.Lbm
    cdef np.float64_t Rh = bm_pars.Rh
    cdef np.float64_t Lh = bm_pars.Lh
    cdef np.ndarray[np.float64_t] ampl_corr = bm_pars.ampl_corr
    cdef np.ndarray[np.float64_t] Abm = bm_pars.Abm

    cdef np.ndarray[np.float64_t] Z13 = np.zeros(100)
    cdef np.ndarray[np.float64_t] Z32 = np.zeros(100)
    cdef np.ndarray[np.float64_t] Z42 = np.zeros(100)
    cdef np.ndarray[np.float64_t] Z43 = np.zeros(100)
    cdef np.ndarray[np.float64_t] g11 = np.zeros(100)
    cdef np.ndarray[np.float64_t] g12 = np.zeros(100)
    cdef np.ndarray[np.float64_t] g2 = np.zeros(100)
    cdef np.ndarray[np.float64_t] g3 = np.zeros(100)
    cdef np.ndarray[np.float64_t] g41 = np.zeros(100)
    cdef np.ndarray[np.float64_t] g42 = np.zeros(100)
    cdef np.ndarray[np.float64_t] a21 = np.zeros(100)
    cdef np.ndarray[np.float64_t] b14 = np.zeros(100)
    cdef np.ndarray[np.float64_t] b44 = np.zeros(100)
    cdef np.ndarray[np.float64_t] b30 = np.zeros(100)
    cdef np.ndarray[np.float64_t] b33 = np.zeros(100)
    cdef np.ndarray[np.float64_t] b20 = np.zeros(100)
    cdef np.ndarray[np.float64_t] b23 = np.zeros(100)

    cdef np.float64_t gh, R_input, Zhel
    cdef np.float64_t R14, R44, G33, G23
    cdef np.float64_t a10, b11, b12, b13, a40, b41, b42, b43
    cdef np.float64_t b21, b22, b31, b32, ah0, bh1, bh2, a14


    cdef Py_ssize_t sec
    cdef Py_ssize_t i

    cdef np.ndarray[np.float64_t, ndim=2] xbm = np.zeros((len(signal), 100))
    cdef np.float64_t sample
    cdef np.ndarray[np.float64_t] time_slice


    cdef Py_ssize_t sections = 100


    # Init gXX
    for sec in range(sections-1,-1,-1):
        if sec == sections-1:
            R14 = Rh + (2*fs*Lh)
            gh = Rh / R14

        # Adaptor 4 (series)
        R44 = Rbm[sec] + 2*fs*Lbm[sec] + 1/(2*fs*Cbm[sec])
        g41[sec] = Rbm[sec] / R44
        g42[sec] = 2*fs*Lbm[sec] / R44

        # Adaptor 3 (parallel)
        G33 = 1/R44 + 2*fs*Ct[sec]
        g3[sec] = 1/(G33*R44)

        # Adaptor 2 (parallel)
        G23 = 1/R14 + G33
        g2[sec] = 1 / (R14*G23)

        # Adaptor 1 (series)
        R14 = 1/G23 + Rs[sec] + 2*fs*Ls[sec]
        g11[sec] = 1 / (G23*R14)
        g12[sec] = Rs[sec] / R14

    R_input = R14
    Zhel = 0

    # Iterate over time
    for i in range(len(signal)):
        sample = signal[i]
        time_slice = xbm[i]

        # Backward wave
        for sec in range(sections-1,-1,-1):
            if sec == sections-1:
                a21[sec] = Zhel
            else:
                a21[sec] = -b14[sec+1]

            b44[sec] = -(-Z42[sec] + Z43[sec])

            b30[sec] = -g3[sec]*(Z32[sec] - b44[sec])
            b33[sec] = Z32[sec] + b30[sec]

            b20[sec] = -g2[sec]*(b33[sec]-a21[sec])
            b23[sec] = b33[sec] + b20[sec]

            b14[sec] = -(b23[sec] - Z13[sec])

        # Forward wave
        for sec in range(sections):
            if sec == 0:
                a14 = 2*R_input*sample + b14[0]
            else:
                a14 = -b21

            a10 = a14 - b14[sec]
            b11 = b23[sec] - g11[sec]*a10
            b12 = -g12[sec]*a10
            b13 = -(b11 + b12 + a14)

            b22 = b11 + b20[sec]
            b21 = b22 + b33[sec] - a21[sec]

            b22 = b11 + b20[sec]
            b21 = b22 + b33[sec] - a21[sec]

            b32 = b22 + b30[sec]
            b31 = b32 + Z32[sec] - b44[sec]

            a40 = b31 - b44[sec]
            b41 = -g41[sec]*a40
            b42 = -Z42[sec] - g42[sec]*a40
            b43 = -(b41 + b42 + b31)

            time_slice[sec] = (b43+Z43[sec])*Cbm[sec]/2/Abm[sec]*ampl_corr[sec]

            Z13[sec] = b13
            Z32[sec] = b32
            Z42[sec] = b42
            Z43[sec] = b43


        # Helicotrema
        ah0 = b21 - Zhel
        bh1 = -gh * ah0
        bh2 = -(bh1 + b21)
        Zhel = bh2


    xbm = np.fliplr(xbm)

    out = dict(zip(bm_pars.real_freq_map, xbm.T))

    return out





def run_lcr4(
        np.ndarray[np.float64_t] xbm,
        np.float64_t fs,
        np.float64_t cf,
):

    assert fs == 48e3

    cdef Py_ssize_t sec = np.where(bm_pars.real_freq_map == cf)[0][0]

    ### The original implementation (c code) had high CFs on the left.
    ### The pre-calculated values in bm_pars have the orignal order
    ### and must be inversed.
    cdef np.ndarray[np.float64_t] freq_map = bm_pars.freq_map[::-1]
    cdef np.ndarray[np.float64_t] Qmin = bm_pars.Qmin[::-1]
    cdef np.ndarray[np.float64_t] Qmax = bm_pars.Qmax[::-1]
    cdef np.ndarray[np.float64_t] SAT1 = bm_pars.SAT1[::-1]
    cdef np.ndarray[np.float64_t] SAT4 = bm_pars.SAT4[::-1]


    cdef np.float64_t T4C, T4L, T5C, T5L, T6C, T6L
    cdef np.float64_t T7C, T7L, Rtest, Rc_res, Rl_res
    cdef np.float64_t Z_CLP1,Z_CLP2,Z_CLP3,Z_CLP4,g_CLP
    cdef np.float64_t p_SAT1,p_SAT2,p_SAT3,p_SAT4
    cdef np.float64_t Qd0, Qd1, Qd2, Qd3

    cdef np.float64_t R1, R2, tau

    cdef np.float64_t V2, V3, V4, V5
    cdef np.float64_t X2, X3, X4, X5
    cdef np.float64_t Rr11, Rr13, Gr21, Gr23, gr1, gr2
    cdef np.float64_t b11, b12, b13, b20, b21, b22, b23, a0

    cdef np.float64_t rect, rect_LP, b2

    cdef Py_ssize_t i

    tau = 1 / (2*np.pi*0.8e3)
    R2 = 1 / (2*fs*tau)
    R1 = 1

    g_CLP = (R2-R1) / (R2+R1)

    Rc_res = Qmin[sec] / (2*fs)
    Rl_res = Rc_res * (fs/(np.pi*freq_map[sec]))**2

    T4C = 0
    T4L = 0
    T5C = 0
    T5L = 0
    T6C = 0
    T6L = 0
    T7C = 0
    T7L = 0

    Qd0 = Qmin[sec]
    Qd1 = Qmin[sec]
    Qd2 = Qmin[sec]
    Qd3 = Qmin[sec]

    p_SAT1 = SAT1[sec]
    p_SAT4 = SAT4[sec]

    p_SAT2 = 10**(np.log10(p_SAT4) + np.log10(p_SAT1/p_SAT4)*2/3)
    p_SAT3 = 10**(np.log10(p_SAT4) + np.log10(p_SAT1/p_SAT4)/3)

    Z_CLP1 = 0
    Z_CLP2 = 0
    Z_CLP3 = 0
    Z_CLP4 = 0

    xbm_out = np.zeros_like(xbm)

    for i in range(len(xbm)):

        ### Resonator 1
        Rr11 = sqrt(Rl_res*Rc_res) / Qd0
        Rr13 = Rr11 + Rl_res

        Gr21 = 1 / Rr13
        Gr23 = Gr21 + 1 / Rc_res

        gr1 = Rr11 / Rr13
        gr2 = Gr21 / Gr23

        ### Filtering
        b13 = -(xbm[i] + T4L)
        b20 = -gr2*(T4C - b13)
        b23 = b20 + T4C

        b22 = b20+b23
        b21 = b22+T4C-b13
        a0 = b21-b13
        b11 = xbm[i]-gr1*a0
        b12 = -(b11+b21)

        X2 = (b22+T4C)/2
        V2 = (T4C-b22)/(2*Rc_res)

        T4L = -b12
        T4C = b22

        rect = 1- (2. / (1+ exp( -p_SAT1*fabs(X2)) ) -1)
        b2 = rect*(1 + g_CLP) - Z_CLP1*g_CLP
        rect_LP = (b2 + Z_CLP1)/2.
        Z_CLP1 = b2

        Qd0 = (Qmax[sec]-Qmin[sec])*rect_LP+Qmin[sec]

        ### Resonator 2
        # Calculating port values from old Qd-value
        Rr11 = sqrt(Rl_res*Rc_res) / Qd1 # sqrt(L/C)/Q
        Rr13 = Rr11 + Rl_res

        Gr21 = 1/Rr13
        Gr23 = Gr21 + 1/Rc_res

        gr1 = Rr11/Rr13
        gr2 = Gr21/Gr23

        ### Filtering
        b13 = -(X2 + T5L)
        b20 = -gr2*(T5C - b13)
        b23 = b20 + T5C

        b22 = b20+b23
        b21 = b22+T5C-b13
        a0 = b21-b13
        b11 = X2-gr1*a0
        b12 = -(b11+b21)

        X3 = (b22+T5C)/2.
        V3 = (T5C-b22)/(2.*Rc_res)

        T5L = -b12
        T5C = b22

        rect = 1- (2. / (1+exp( -p_SAT2*fabs(X3)) ) -1)
        b2 = rect*(1 + g_CLP) - Z_CLP2*g_CLP
        rect_LP = (b2 + Z_CLP2)/2.
        Z_CLP2 = b2

        Qd1 = (Qmax[sec]-Qmin[sec])*rect_LP+Qmin[sec]


        ### Resonator 3
        Rr11 = sqrt(Rl_res*Rc_res)/Qd2
        Rr13 = Rr11 + Rl_res

        Gr21 = 1/Rr13
        Gr23 = Gr21 + 1/Rc_res

        gr1 = Rr11/Rr13
        gr2 = Gr21/Gr23

        ### Filtering
        b13 = -(X3+T6L)
        b20 = -gr2*(T6C-b13)
        b23 = b20+T6C

        b22 = b20+b23
        b21 = b22+T6C-b13
        a0 = b21-b13
        b11 = X3-gr1*a0
        b12=-(b11+b21)

        X4 = (b22+T6C)/2.
        V4 = (T6C-b22)/(2.*Rc_res)


        T6L = -b12
        T6C = b22

        rect = 1- (2. / (1+exp( -p_SAT3*fabs(X4)) ) -1)
        b2 = rect*(1 + g_CLP) - Z_CLP3*g_CLP
        rect_LP = (b2 + Z_CLP3)/2.
        Z_CLP3 = b2

        Qd2 = (Qmax[sec]-Qmin[sec])*rect_LP+Qmin[sec]

        ### Resonator 4
        Rr11 = sqrt(Rl_res*Rc_res)/Qd3
        Rr13 = Rr11 + Rl_res

        Gr21 = 1/Rr13
        Gr23 = Gr21 + 1/Rc_res

        gr1 = Rr11/Rr13
        gr2 = Gr21/Gr23

        ### Filtering
        b13 = -(X4+T7L)
        b20 = -gr2*(T7C-b13)
        b23 = b20+T7C

        b22 = b20+b23
        b21 = b22+T7C-b13
        a0 = b21-b13
        b11 = X4-gr1*a0
        b12 = -(b11+b21)

        X5 = (b22+T7C)/2.
        V5 = (T7C-b22)/(2.*Rc_res)

        T7L = -b12
        T7C = b22

        rect = 1 - (2 / (1 + exp( -p_SAT4*fabs(X5)) ) - 1)
        b2 = rect*(1 + g_CLP) - Z_CLP4*g_CLP
        rect_LP = (b2 + Z_CLP4) / 2
        Z_CLP4 = b2

        Qd3 = (Qmax[sec]-Qmin[sec])*rect_LP+Qmin[sec]



        xbm_out[i] = X5

    return xbm_out






def run_ihcrp(
        np.ndarray[np.float64_t] xbm,
        np.float64_t fs,
        np.float64_t cf,
):

    assert fs == 48e3

    cdef Py_ssize_t sec = np.where(bm_pars.real_freq_map == cf)[0][0]
    sec = 99 - sec         # Flipping the sections, because the
                           # original C implementation had high CFs on
                           # the left hand side

    cdef np.ndarray[np.float64_t] uIHC = np.zeros_like(xbm)
    cdef np.ndarray[np.float64_t] ciliaCouplingGain = bm_pars.ciliaGain

    cdef np.float64_t p_endocochlearPot_Et = 0.1
    cdef np.float64_t p_reversalPot_Ek = -0.07045
    cdef np.float64_t p_reversalPotCorrection = 0.04
    cdef np.float64_t p_totalCapacitance_C = 6e-12
    cdef np.float64_t p_restingConductance_G0 = 1.974e-09
    cdef np.float64_t p_kConductance_Gk = 1.8e-08
    cdef np.float64_t p_maxMConductance_Gmax = 8e-09
    cdef np.float64_t p_ciliaTimeConst_tc = 0.00213
    cdef np.float64_t p_referencePot = 0.0
    cdef np.float64_t p_sensitivity_s0 = 85e-09
    cdef np.float64_t p_sensitivity_s1 = 5e-09
    cdef np.float64_t p_offset_u0 = 7e-09
    cdef np.float64_t p_offset_u1 = 7e-09

    cdef np.float64_t p_C_ST = 4.0 # IHC bundle compliance in m/N
    cdef np.float64_t p_F0_ST = 1.e3 # IHC bundle stiffness-fluid friction corner frequency to calculate R_ST (Ns/m)

    cdef np.float64_t Z_ST, g_ST, uIHC_old
    cdef np.float64_t dtOverC, gkEpk
    cdef np.float64_t restingPotential_V0

    cdef np.float64_t L_ST, R_ST, f0_ST
    cdef np.float64_t dt

    cdef np.float64_t leakageConductance_Ga,
    cdef np.float64_t conductance_G
    cdef np.float64_t potential_V
    cdef np.float64_t ciliaAct
    cdef np.float64_t u0, u1, s0, s1
    cdef np.float64_t b1_ST, b2_ST, b3_ST
    cdef np.float64_t xST

    cdef Py_ssize_t i

    cdef int sections = 100

    restingPotential_V0 = (
        (
            p_restingConductance_G0*p_endocochlearPot_Et
            + p_kConductance_Gk*(p_reversalPot_Ek + p_endocochlearPot_Et*p_reversalPotCorrection)
        )
        /
        (p_restingConductance_G0 + p_kConductance_Gk)
    )


    dt = 1 / fs
    dtOverC = dt / p_totalCapacitance_C
    gkEpk = (
        p_kConductance_Gk
        *
        (p_reversalPot_Ek + p_endocochlearPot_Et*p_reversalPotCorrection)
    )

    Z_ST = 0

    L_ST = 1 / p_C_ST            # v-U / F-I analogy !!!
    f0_ST = 2000 * (( (2000/200)**(1/sections) )**(-sec)) # grade Fluid filter from 2000 to 200 Hz

    R_ST = L_ST * 2 * np.pi * f0_ST     # Fluid friction on bundles Ns/m

    g_ST = R_ST / (R_ST + 2*fs*L_ST)  # gamma

    s0 = p_sensitivity_s0
    u0 = p_offset_u0
    s1 = p_sensitivity_s1
    u1 = p_offset_u1


    ciliaAct = 1.0 / (1.0 + exp(u0 / s0) * ( 1 + exp(u1 / s1)))
    leakageConductance_Ga = (p_restingConductance_G0 -
                             p_maxMConductance_Gmax * ciliaAct)

    uIHC_old = restingPotential_V0

    for i in range(len(xbm)):
        b3_ST = -(xbm[i] + Z_ST)
        b1_ST = xbm[i] + 2*g_ST*b3_ST
        b2_ST = -(2*xbm[i] + 2*g_ST*b3_ST + Z_ST)
        xST = -(b2_ST+Z_ST)/2

        Z_ST = -b2_ST


        xST *= ciliaCouplingGain[sec]


        potential_V = uIHC_old

        ciliaAct = 1 / (1 + exp((u0 - xST) / s0) *
                        (1 + exp((u1 - xST) / s1)))

        conductance_G = (p_maxMConductance_Gmax * ciliaAct +
                         leakageConductance_Ga)


        uIHC[i] = (potential_V - dtOverC *
                   (conductance_G * (potential_V - p_endocochlearPot_Et) +
                    p_kConductance_Gk * potential_V - gkEpk))


        uIHC_old = uIHC[i]

    return uIHC




def run_ihc_meddis2000(
        np.ndarray[np.float64_t, ndim=1] ihcrp,
        np.float64_t fs,
        np.float64_t gamma_ca,
        np.float64_t beta_ca,
        np.float64_t tau_m,
        np.float64_t tau_ca,
        np.float64_t gmax_ca,
        np.float64_t e_ca,
        np.float64_t perm_ca0,
        np.float64_t perm_z,
        np.float64_t power_ca,
        np.float64_t replenish_rate_y,
        np.float64_t loss_rate_l,
        np.float64_t recovery_rate_r,
        np.float64_t reprocess_rate_x,
        np.float64_t max_free_pool,
        syn_mode='probability'
):

    cdef np.float64_t dt = 1/fs
    cdef np.ndarray[np.float64_t, ndim=1] psp
    cdef np.float64_t uIHC_rest
    cdef np.float64_t conc_Ca
    cdef np.float64_t act_Ca_inf
    cdef np.float64_t k0        # spont perm
    cdef np.float64_t vin       # resting potential
    cdef np.float64_t act_Ca
    cdef np.float64_t reservoirQ
    cdef np.float64_t spontFreePool_q0
    cdef np.float64_t cleftC
    cdef np.float64_t spontCleft_c0
    cdef np.float64_t ICa
    cdef np.float64_t replenish
    cdef np.float64_t ejected
    cdef np.float64_t kdt
    cdef np.float64_t reUptakeAndLost
    cdef np.float64_t reUptake
    cdef np.float64_t reprocessed
    cdef np.float64_t reprocessedW
    cdef Py_ssize_t i



    ### Initial values
    psp = np.zeros_like(ihcrp)
    uIHC_rest = ihcrp[0]

    act_Ca_inf = 1 / (1 + exp(-gamma_ca*uIHC_rest)/beta_ca)
    ICa = gmax_ca * act_Ca_inf**3 * (uIHC_rest - e_ca)

    if -ICa > perm_ca0:
        k0 = perm_z * ((-ICa)**power_ca - perm_ca0**power_ca)
    else:
        k0 = 0.0

    # cleftReplenishMode == IHC_MEDDIS2000_CLEFTREPLENISHMODE_ORIGINAL
    spontCleft_c0 = (
        max_free_pool * replenish_rate_y * k0
        /
        (replenish_rate_y * (loss_rate_l+recovery_rate_r) + k0*loss_rate_l)
    )

    if (spontCleft_c0 > 0) and syn_mode == 'probability':
        spontFreePool_q0 = spontCleft_c0 * (loss_rate_l + recovery_rate_r) / k0
    elif (spontCleft_c0 > 0) and syn_mode == 'quantal':
        spontFreePool_q0 = np.floor( (spontCleft_c0 * (loss_rate_l+recovery_rate_r) / k0) + 0.5 )
    else:
        spontFreePool_q0 = max_free_pool

    spontReprocess_w0 = spontCleft_c0 * recovery_rate_r / reprocess_rate_x

    act_Ca = act_Ca_inf
    conc_Ca = -ICa
    reservoirQ = spontFreePool_q0
    cleftC = spontCleft_c0
    reprocessedW = spontReprocess_w0


    for i in range(len(ihcrp)):

        vin = ihcrp[i]

        ### Ca current
        act_Ca_inf = 1 / (1 + exp(-gamma_ca*vin)/beta_ca)
        act_Ca += (act_Ca_inf - act_Ca) * dt / tau_m
        ICa = gmax_ca * act_Ca**3 * (vin - e_ca)


        ### Calcium Ion accumulation and diffusion
        # caCondMode == IHC_MEDDIS2000_CACONDMODE_ORIGINAL
        conc_Ca += (-ICa - conc_Ca) * dt / tau_ca


        ### power law release function
        if conc_Ca > perm_ca0:
            kdt = (perm_z*dt * (conc_Ca**power_ca - perm_ca0**power_ca))
        else:
            kdt = 0.0



        ### Synapse
        if syn_mode == 'probability':
            # cleftReplenishMode == IHC_MEDDIS2000_CLEFTREPLENISHMODE_ORIGINAL
            if reservoirQ < max_free_pool:
                replenish = replenish_rate_y * dt * (max_free_pool - reservoirQ)
            else:
                replenish = 0.0


            ejected = kdt * reservoirQ

            reUptakeAndLost = (loss_rate_l + recovery_rate_r) * dt * cleftC
            reUptake = recovery_rate_r * dt * cleftC

            reprocessed = reprocess_rate_x * dt * reprocessedW
            reservoirQ += replenish - ejected + reprocessed
            cleftC += ejected - reUptakeAndLost

            psp[i] = ejected

            reprocessedW += reUptake - reprocessed

        elif syn_mode == 'quantal':
            if reservoirQ < max_free_pool:
                replenish = (np.random.geometric(
                    replenish_rate_y * dt,
                    int(max_free_pool - reservoirQ)
                ) == 1).sum()
            else:
                replenish = 0.0

            ejected = (np.random.geometric(
                kdt,
                int(reservoirQ)
            ) == 1).sum()

            reUptakeAndLost = (loss_rate_l+recovery_rate_r)*dt * cleftC
            reUptake = recovery_rate_r * dt * cleftC
            if reprocessedW < 1:
                reprocessed = 0.0
            else:
                reprocessed = (np.random.geometric(
                    reprocess_rate_x * dt,
                    int(reprocessedW)
                ) == 1).sum()

            reservoirQ += replenish - ejected + reprocessed
            cleftC += ejected - reUptakeAndLost


            if ejected > 0:
                psp[i] = ejected
            else:
                psp[i] = 0.0

            reprocessedW += reUptake - reprocessed


        else:
            raise NotImplementedError("Unimplemented syn_mode: {}".format(syn_mode))

    return psp



def run_an_sg_carney_holmberg2007(
        np.ndarray[np.float64_t, ndim=1] psp,
        np.float64_t fs,
        np.float64_t c0,
        np.float64_t c1,
        np.float64_t s0,
        np.float64_t s1,
        np.float64_t refractory_period,
):

    cdef np.float64_t dt
    cdef np.float64_t timer
    cdef Py_ssize_t i
    cdef np.float64_t p
    cdef np.float64_t excess_time
    cdef np.float64_t threshold
    cdef np.ndarray[np.float64_t, ndim=1] rand

    dt = 1/fs
    timer = refractory_period + dt
    rand = np.random.rand(len(psp))

    spikes = []
    for i in range(len(psp)):

        if timer > refractory_period:
            excess_time = timer - refractory_period
            threshold = (
                c0*exp(-excess_time/s0)
                +
                c1*exp(-excess_time/s1)
            )

            if (1-threshold)*psp[i] > rand[i]:
                spikes.append(i*dt)
                timer = 0

        timer += dt

    spikes = np.array(spikes)
    return spikes



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo calculates rate based hearing threshold of an innear ear
model with automatic speach prefiltering in order to adjust thresholds
to human hearing thresholds.

On a fast computer it will take around 1 min to finish.  If you set
shell variable THmap=m, then it will use multiple cores.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea.external
from cochlea.stats import calc_thresholds_rate, calc_human_hearing_thresholds


def main():

    cfs = np.logspace(np.log10(125), np.log10(16e3), 32)

    ths = calc_thresholds_rate(
        model=cochlea.run_zilany2014,
        cfs=cfs,
        model_pars={'species': 'human'},
        asr_filter=True
    )


    # ths = calc_thresholds_rate(
    #     model=cochlea.external.run_matlab_auditory_periphery,
    #     cfs=cfs,
    #     model_pars={},
    #     asr_filter=True
    # )


    human_ths = calc_human_hearing_thresholds(cfs)

    fig, ax = plt.subplots()

    human_ths.plot(ax=ax, logx=True, style='--', linewidth=5)
    ths.plot(ax=ax, logx=True)

    plt.show()


if __name__ == "__main__":
    main()



"""Modulation gain of inner ear models.

"""
from __future__ import division, print_function, absolute_import

import numpy as np

import thorns as th
import thorns.waves as wv

from . threshold_rate import calc_thresholds_rate

import logging

log = logging.getLogger('thorns')


def calc_modulation_gain(
        model,
        fms=None,
        cf=10e3,
        model_pars=None,
        m=1,
        level_above_threshold=10,
        map_backend=None,
):
    """Calculate modulation gain of an inner ear model.

    Parameters
    ----------
    fms : array_like
        List of modulation frequencies (Hz).
    cf : scalar, optional
        Characteristic frequency (Hz).
    model_pars : dict, ooptional
        Additional parameters to be passed to the model funtion.
    m : float, optional
        Modulation depth in the range of <0, 1>.
    level_above_threshold : scalar, optional
        Sound level which will be added to the threshold level to
        calculate modulation gain.

    map_backend : str, optional
        Map backend that will be used in the calculation.  See
        `thorns.util.map()`.

    """
    if model_pars is None:
        model_pars = {}

    if fms is None:
        fms = np.logspace(np.log10(10), np.log10(2e3), 16)

    # Calculate dbspl = threshold + 10 dB
    threshold = calc_thresholds_rate(
        model=model,
        cfs=[cf],
        model_pars=model_pars
    )

    dbspl = threshold['threshold'].iloc[0] + level_above_threshold
    log.info("Sound level: {} dB SPL".format(dbspl))

    space = {
        'fm': fms,
    }

    kwargs = {
        'model': model,
        'cf': cf,
        'dbspl': dbspl,
        'model_pars': model_pars,
        'm': m,
    }

    gains = th.util.map(
        _run_model,
        space,
        kwargs=kwargs,
        backend=map_backend,
    )

    gains.columns = ['gain']

    return gains


def _run_model(model, fm, cf, dbspl, model_pars, m):

    duration = 0.6
    onset = 10e-3

    fs = model_pars.setdefault('fs', 100e3)
    model_pars.setdefault('anf_num', (250, 0, 0))
    model_pars.setdefault('seed', 0)

    sound = wv.amplitude_modulated_tone(
        fs=fs,
        fm=fm,
        fc=cf,
        m=m,
        duration=duration,
        dbspl=dbspl,
    )

    anf = model(
        sound=sound,
        cf=cf,
        **model_pars
    )

    trimmed = th.trim(anf, onset)

    vs = th.vector_strength(trimmed, fm)
    gain = 20 * np.log10(2*vs / m)

    return gain



#!/usr/bin/env python
"""
Run inner ear model from [Zilany2009]_.

"""
from __future__ import division, absolute_import, print_function


import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as dsp

import thorns as th
import cochlea

def main():

    fs = 100e3

    # Make sound
    t = np.arange(0, 0.1, 1/fs)
    s = dsp.chirp(t, 80, t[-1], 20000)
    s = cochlea.set_dbspl(s, 50)
    s = np.concatenate((s, np.zeros(int(10e-3*fs))))

    # Run model
    anf = cochlea.run_zilany2009(
        s,
        fs,
        anf_num=(100,0,0),
        cf=(80, 20000, 100),
        seed=0,
        powerlaw='approximate'
    )

    # Plot auditory nerve response
    anf_acc = th.accumulate(anf, keep=['cf', 'duration'])
    anf_acc.sort_values('cf', ascending=False, inplace=True)

    cfs = anf.cf.unique()

    fig, ax = plt.subplots(2,1, sharex=True)
    th.plot_neurogram(
        anf_acc,
        fs,
        ax=ax[0]
    )

    th.plot_raster(
        anf[anf.cf==cfs[30]],
        ax=ax[1]
    )

    ax[1].set_title("CF = {}".format(cfs[30]))

    plt.show()



if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division


raise NotImplementedError


import numpy as np
import os
import multiprocessing

import thorns.waves as wv
import thorns as th

def _run_model( (model, cf, fs, freq, dbspl, kwargs) ):
    print os.getpid(), freq, dbspl

    tmax = 250
    onset = 30

    ear = model((1000, 1000, 1000), cf=cf, **kwargs)
    # ear = model((250, 0, 0), cf=cf,
    #             powerlaw_implnt='approx',
    #             with_ffGn=False)

    s = wv.generate_ramped_tone(fs, freq,
                                tone_duration=tmax,
                                pad_duration=0,
                                dbspl=dbspl)
    anf = ear.run(fs, s)

    hsr = anf.where(typ='hsr')
    hsr = th.trim(hsr, onset)
    hsr_rate = th.calc_rate(hsr, stimulus_duration=(tmax-onset))

    msr = anf.where(typ='msr')
    msr = th.trim(msr, onset)
    msr_rate = th.calc_rate(msr, stimulus_duration=(tmax-onset))

    lsr = anf.where(typ='lsr')
    lsr = th.trim(lsr, onset)
    lsr_rate = th.calc_rate(lsr, stimulus_duration=(tmax-onset))

    return freq, dbspl, hsr_rate, msr_rate, lsr_rate


def calc_isointensity_curves(model,
                             cf=3000,
                             fs=100e3,
                             freqs=np.logspace(np.log10(500), np.log10(6000), 32),
                             dbspls=np.arange(0, 100, 10),
                             **kwargs):

    space = [(model, cf, fs, freq, dbspl, kwargs)
             for m in [model]
             for cf in [cf]
             for fs in [fs]
             for freq in freqs
             for dbspl in dbspls
             for kwargs in [kwargs]]

    pool = multiprocessing.Pool()
    rates = pool.map(_run_model, space)

    rates = np.rec.array(rates, names='freq,dbspl,hsr_rate,msr_rate,lsr_rate')

    return rates


def main():
    # import pycat
    # model = pycat.Zilany2009
    # pars = { 'powerlaw_implnt':'approx',
    #          'with_ffGn':False }
    # fs = 100e3

    import cochlea
    model = cochlea.Sumner2003
    pars = {}
    fs = 100e3

    # import cochlea
    # model = cochlea.Holmberg2007
    # fs = 48000


    import traveling_waves as tw
    cf = tw.real_freq_map[68]


    # print _run_model( (model, cf, 48000, 3000, 50, {}) )
    # exit()

    rates = calc_isointensity_curves(model, cf,
                                     fs=fs,
                                     **pars)
    print rates

    import biggles
    p = biggles.FramedPlot()
    p.xlog = 1
    for dbspl in np.unique(rates['dbspl']):
        selected = rates[ rates['dbspl']==dbspl ]
        p.add( biggles.Curve(selected['freq'], selected['hsr_rate']) )

    p.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
import scipy.io

import os
from os.path import join

import cochlea.zilany2014._zilany2014 as zilany2014
from cochlea.zilany2014.util import ffGn

DATADIR = os.path.join(
    os.path.dirname(__file__),
    'data_zilany2014'
)

def test_ihc():

    m = scipy.io.loadmat(
        join(DATADIR, 'data_zilany2014.mat'),
        squeeze_me=True
    )
    fs = float(m['fs'])
    cf = float(m['cf'])
    sound = m['sound'].astype(float)
    vihc_target = m['vihc']

    vihc = zilany2014.run_ihc(
        signal=sound,
        cf=cf,
        fs=fs,
        species='cat',
        cohc=1.,
        cihc=1.
    )

    assert_almost_equal(
        vihc,
        vihc_target,
        decimal=15
    )




def test_synapse():
    """test_synapse()

    This function has problems, because it's using matlab
    implementation of `resample' and it's difficult to exactly
    reproduce the output.

    During the generation of mat-file the `ffGn' was replaced by
    `zeros' and Matlab's `resample' by self-made downsampling function
    equivalent to python's implementation:

    function resampled = resample_fake(X, P, Q)
        b = fir1(Q, 1/Q);
        a = [1];
        filtered = filtfilt(b, a, X);
        resampled = filtered(1:Q:end);

    """
    m = scipy.io.loadmat(
        join(DATADIR, 'data_zilany2014.mat'),
        squeeze_me=True
    )
    fs = float(m['fs'])
    cf = float(m['cf'])
    vihc = m['vihc']
    meanrate_target = m['meanrate']

    synout = zilany2014.run_synapse(
        vihc=vihc,
        fs=fs,
        cf=cf,
        anf_type='hsr',
        powerlaw='approximate',
        ffGn=False
    )
    meanrate = synout / (1 + 0.75e-3*synout)

    assert_almost_equal(
        meanrate,
        meanrate_target,
        decimal=10
    )


def test_ffGn():
    m = scipy.io.loadmat(
        join(DATADIR, 'data_zilany2014.mat'),
        squeeze_me=True
    )

    r = m['random_ffGn']
    y = m['y_ffGn']

    y_actual = ffGn(
        N=16,
        tdres=0.1,
        Hinput=0.2,
        noiseType=1,
        mu=10,
        random_debug=r
    )

    assert_almost_equal(y_actual, y, decimal=13)



# -*- coding: utf-8 -*-

"""Interoperability with Brian.

"""
from __future__ import division, print_function, absolute_import

import numpy as np

import cochlea
import thorns as th
import thorns.waves as wv

import brian
from brian import mV, second, ms, volt

def main():

    fs = 100e3
    cf = 1e3
    tmax = 50e-3

    sound = wv.ramped_tone(
        fs=fs,
        freq=cf,
        duration=tmax,
        pad=30e-3,
        dbspl=50,
    )

    anf_trains = cochlea.run_zilany2014(
        sound=sound,
        fs=fs,
        cf=cf,
        anf_num=(300, 0, 0),
        seed=0,
        species='cat',
    )

    anfs = cochlea.make_brian_group(anf_trains)

    print(anfs)


    brainstem = make_brainstem_group(100)

    print(brainstem)

    monitor = brian.SpikeMonitor(brainstem)


    net = brian.Network([brainstem, monitor])
    net.run(tmax*second)


    brian.raster_plot(monitor)
    brian.show()



def make_brainstem_group(num):
    """Generate integrate-and-fire neuron group that can be connected to a
    auditory nerve fibers group.

    """
    eqs = '''
    dv/dt = (ge-(v+49*mV))/(2*ms) : volt
    dge/dt = -ge/(0.5*ms) : volt
    '''

    group = brian.NeuronGroup(num, eqs, threshold=-50*mV, reset=-60*mV)
    group.v = -60*mV

    return group



if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""This example illustrates how to run Zilany et al. (2014) model with
the rate (not spikes) as output.

Note that this functionality is given for convenience and not very
well supported.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as dsp

import thorns as th

import cochlea

def main():

    fs = 100e3

    ### Make sound
    t = np.arange(0, 0.1, 1/fs)
    s = dsp.chirp(t, 80, t[-1], 20000)
    s = cochlea.set_dbspl(s, 50)
    s = np.concatenate( (s, np.zeros(int(10e-3 * fs))) )



    ### Run model
    rates = cochlea.run_zilany2014_rate(
        s,
        fs,
        anf_types=['msr'],
        cf=(125, 20000, 100),
        powerlaw='approximate',
        species='human'
    )


    ### Plot rates
    fig, ax = plt.subplots()
    img = ax.imshow(
        rates.T,
        aspect='auto'
    )
    plt.colorbar(img)
    plt.show()



if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from __future__ import division, absolute_import, print_function

import numpy as np
import pandas as pd

import matlab_wrapper

import thorns as th


def run_matlab_auditory_periphery(
        sound,
        fs,
        anf_num,
        cf,
        seed,
        params_name='Normal',
        matlab_session=None
):
    """Run Matlab Auditory Periphery [MAP]_ model by Ray Meddis.  This
    function does not implement the model, but wraps the model
    implementation using `matlab_wrapper`.  The model takes sound
    signal as input and outputs auditory nerve spike trains.

    In order to run it, make sure that all necessary [MAP]_ model
    files are in `MATLABPATH`.  You should be able to run `MAP1_14`
    function in MATLAB first!

    Requires MATLAB, *matlab_wrapper*, *thorns* and [MAP]_ in
    `MATLABPATH`.


    Parameters
    ----------
    sound : array_like
        Input sound.
    fs : float
        Sampling frequency of the sound.
    anf_num : tuple
        Number of auditory nerve fibers per channel (HSR#, MSR#, LSR#).
    cf : float or array_like or tuple
        The center frequency(s) of the simulated auditory nerve fibers.
        If float, then defines a single frequency channel.  If
        array_like (e.g. list or ndarray), then the frequencies are
        used.  If tuple, then must have exactly 3 elements (min_cf,
        max_cf, num_cf) and the frequencies are calculated using the
        Greenwood function.
    seed : int
        Random seed.
    params_name : str, optional Tail of the parameter filename
        (parameterStore/MAPparams<params_name>.m).  Refer to MAP
        documentation for the details.
    matlab_session : MatlabSession or None, optional
        MatlabSession object from `matlab_wrapper` module.  If `None`,
        then new session is generated.


    Returns
    -------
    spike_trains
        Auditory nerve spike trains.


    References
    ----------

    .. [MAP] http://www.essexpsychology.macmate.me/HearingLab/modelling.html

    """

    ### Validate `anf_num`
    assert len(anf_num) == 3


    ### Validate `cf`
    if isinstance(cf, tuple):
        assert len(cf) == 3
    elif np.isscalar(cf):
        pass
    elif len(cf) == 3:
        raise RuntimeError("Three frequency channels are forbidden, because they mask the tuple (min_cf, max_cf, cf_num).")


    ### Generate matlab_wrapper session as needed
    if matlab_session is None:
        matlab = matlab_wrapper.MatlabSession(options='-nosplash -singleCompThread')
    else:
        matlab = matlab_session

    matlab.eval("clear all")
    matlab.eval("clearvars -global")

    ### Set Matlab environment
    matlab.workspace.rng(seed)

    matlab.eval("global dtSpikes ANoutput savedBFlist")

    matlab.workspace.MAP1_14(
        sound,
        float(fs),
        np.array(cf, dtype=float),
        params_name,
        'spikes',
        ['AN_IHCsynapseParams.numFibers={};'.format(max(anf_num)),
         'AN_IHCsynapseParams.spikesTargetSampleRate={};'.format(fs)],
        nout=0
    )


    ### Collect results from Matlab
    anf_raw = matlab.workspace.ANoutput

    cf_raw = matlab.workspace.savedBFlist
    cf_raw = np.atleast_1d(cf_raw)

    dt_raw = matlab.workspace.dtSpikes



    ### Make trains
    all_trains = th.make_trains(
        anf_raw.T,
        fs=1/dt_raw
    )

    cf_col = np.tile(np.repeat(cf_raw,anf_num[0]), 3)
    type_col = np.repeat(['lsr', 'msr', 'hsr'], len(cf_raw)*anf_num[0])

    all_trains['type'] = type_col
    all_trains['cf'] = cf_col



    n = {
        'hsr': anf_num[0],
        'msr': anf_num[1],
        'lsr': anf_num[2],
    }

    ### Discard trains.  We want only anf_num == (HSR#, MSR#, LSR#)
    anf_trains = []
    for name,group in all_trains.groupby(['type','cf']):
        typ,cf = name

        sel = group.iloc[0:n[typ]]

        anf_trains.append(sel)


    anf_trains = pd.concat(anf_trains)

    return anf_trains



"""Rate based threshold for inner ear models.

"""
from __future__ import division, print_function, absolute_import

import numpy as np
import logging

import thorns as th
import thorns.waves as wv

from cochlea.asr import adjust_to_human_thresholds


def calc_thresholds_rate(
        model,
        cfs=None,
        model_pars=None,
        asr_filter=False,
        map_backend=None,
):
    """Calculate rate based hearing threshold of an inner ear model."""

    if cfs is None:
        cfs = np.logspace(np.log10(125), np.log10(16000), 32)

    if model_pars is None:
        model_pars = {}

    # Calculate spontaneous rate for reference
    spont_rate = th.util.cache(calc_spont_threshold)(
        model=model,
        cf=cfs[0],
        model_pars=model_pars
    )

    space = {
        'cf': cfs
    }
    kwargs = {
        'model': model,
        'model_pars': model_pars,
        'spont_rate': spont_rate,
        'asr_filter': asr_filter,
    }

    thresholds = th.util.map(
        calc_threshold,
        space,
        kwargs=kwargs,
        backend=map_backend,
    )

    thresholds.columns = ['threshold']

    return thresholds


def calc_spont_threshold(model, cf, model_pars):

    pars = dict(model_pars)

    fs = pars.setdefault('fs', 100e3)
    pars.setdefault('seed', 0)
    pars.setdefault('anf_num', (1000, 0, 0))

    tmax = 250e-3
    silence = np.zeros(int(fs*tmax))

    anf = model(
        sound=silence,
        cf=cf,
        **pars
    )

    rates = anf.apply(th.firing_rate, axis=1)

    threshold = rates.mean() + rates.std()

    logging.debug("mean + std rate = {}".format(threshold))

    return threshold


def calc_threshold(
        model,
        cf,
        spont_rate,
        model_pars,
        asr_filter=False,
        freq=None
):

    kwargs = {
        'model': model,
        'model_pars': model_pars,
        'cf': cf,
        'spont_rate': spont_rate,
        'asr_filter': asr_filter,
        'freq': freq,
    }

    dbspl_opt = th.util.find_zero(
        error_func,
        kwargs=kwargs,
        x1=-100,
        x2=100,
        xtol=0.1
    )

    return dbspl_opt


def error_func(
        dbspl,
        model,
        cf,
        spont_rate,
        model_pars,
        asr_filter=False,
        freq=None
):

    pars = dict(model_pars)

    fs = pars.setdefault('fs', 100e3)
    pars.setdefault('seed', 0)
    pars.setdefault('anf_num', (1000, 0, 0))

    if freq is None:
        freq = cf

    tone_duration = 250e-3
    onset = 15e-3

    sound = wv.ramped_tone(
        fs,
        freq,
        duration=tone_duration,
        ramp=2.5e-3,
        pad=0,
        dbspl=dbspl
    )

    if asr_filter:
        sound = adjust_to_human_thresholds(sound, fs, model)

    anf = model(
        sound=sound,
        cf=cf,
        **pars
    )

    trains = th.trim(anf, onset, None)
    rate = th.firing_rate(trains)

    error = rate - spont_rate

    logging.debug("{} {} {}".format(dbspl, rate, error))

    return error



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Automatic Speach Recognition related functions.

"""

from __future__ import division, absolute_import, print_function

import numpy as np
import pandas as pd
import scipy

import inspect
import os


def adjust_to_human_thresholds(signal, fs, model):
    """Filter `signal` in order to adjust threshold of `model` to human
    hearing thresholds.

    """

    if inspect.isfunction(model):
        name = model.__name__
    else:
        name = str(model)


    if 'zilany2014' in name:
        fname = _data_dir("zilany2014.csv")
    elif 'holmberg2007' in name:
        fname = _data_dir("holmberg2007.csv")
    elif 'matlab_auditory_periphery' in name:
        fname = _data_dir("matlab_auditory_periphery.csv")
    else:
        raise NotImplementedError


    filt = pd.read_csv(fname)

    signal_fft = np.fft.rfft(signal)
    freqs = np.linspace(0, fs/2, len(signal_fft))


    # Interpolate the filter to fit signal's FFT
    h = scipy.interpolate.interp1d(
        x=filt.freq,
        y=filt.h,
        kind='cubic',
        bounds_error=False,
        fill_value=-100
    )



    # Convert dB to amplitude ratio
    h_ratio = 10 ** (h(freqs) / 20)


    # Apply the filter
    signal_fft *= h_ratio


    # Go back to time domain
    filtered = np.fft.irfft(signal_fft)


    return filtered










def _data_dir(par_file):
    base_dir = os.path.dirname(__file__)
    return os.path.join(base_dir, par_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""External models.

"""

from __future__ import division, absolute_import, print_function

import warnings

try:
    from . matlab_auditory_periphery import run_matlab_auditory_periphery
except ImportError:
    warnings.warn("Error importing run_matlab_auditory_periphery().  Probably matlab_wrapper not installed.")



from __future__ import division, print_function, absolute_import

import numpy as np
import pandas as pd
import itertools

from . import traveling_waves as tw
from . traveling_waves import real_freq_map, get_nearest_cf

from . holmberg2007_vesicles import run_holmberg2007_vesicles


def run_holmberg2007(
        sound,
        fs,
        anf_num,
        seed,
        cf=None,
        syn_mode='probability'
):
    """Run the inner ear model by [Holmberg2007]_.  It simulates the
    traveling wave on the basilar membrane, inner hair cell, synapses
    and generates auditory nerve spikes.  The model takes sound signal
    as input and outputs auditory nerve spike trains.


    Parameters
    ----------
    sound : array_like
        Input sound signal.
    fs : float
        Sampling frequency of the sound.
    anf_num : tuple
        Number of auditory nerve fibers per channel (HSR#, MSR#, LSR#).
    seed : int
        Random seed.
    cf : float or array_like or None, optional
        Characteristic frequencies.  If `None`, then calculate all 100
        predefined CFs.  CFs must be a subset of `real_freq_map`.
    syn_mode : {'probability', 'quantal'}, optional
        Output mode of the IHC synapse.


    Returns
    -------
    spike_trains
        Auditory nerve spike trains.


    References
    ----------
    If you are using results of this or modified version of the model
    in your research, please cite [Holmberg2007]_.

    .. [Holmberg2007] Holmberg, M. (2007). Speech Encoding in the
       Human Auditory Periphery: Modeling and Quantitative Assessment
       by Means of Automatic Speech Recognition. PhD thesis, Technical
       University Darmstadt.

    """

    assert np.max(np.abs(sound)) < 1000, "Signal should be given in Pa"
    assert sound.ndim == 1
    assert fs == 48e3

    np.random.seed(seed)

    if cf is None:
        cfs = tw.real_freq_map
    elif np.isscalar(cf):
        cfs = [cf]
    else:
        cfs = cf


    assert set(cfs) <= set(tw.real_freq_map), set(cfs) - set(tw.real_freq_map)


    duration = len(sound) / fs



    ### Outer ear filter
    sound_oe = tw.run_outer_ear_filter(sound, fs)

    ### Middle ear filter
    sound_me = tw.run_middle_ear_filter(sound_oe, fs)

    ### Scaling
    sound_scaled = sound_me * tw.scaling_factor

    ### Basilar membrane
    xbm = tw.run_bm_wave(sound_scaled, fs)


    ihcrp = {}
    for cf in cfs:
        ### Amplification
        lcr4 = tw.run_lcr4(xbm[cf], fs, cf)

        ### Delay correction (1/cf)
        sec = np.where(tw.real_freq_map == cf)[0][0]
        lcr4_rolled = np.roll(
            lcr4,
            -int(np.round(tw.delay_time[99-sec]*fs))
        )

        ### IHCRP
        ihcrp[cf] = tw.run_ihcrp(lcr4_rolled, fs, cf)



    anf_types = np.repeat(['hsr', 'msr', 'lsr'], anf_num)


    ihc_meddis2000_pars = {
        'hsr': {                # H2 fiber from Sumner et al. (2002)
            'gamma_ca': 130,
            'beta_ca': 400,
            'tau_m': 1e-4,
            'gmax_ca': 4.5e-9,
            'e_ca': 0.066,
            'tau_ca': 1e-4,
            'perm_ca0': 0,
            'perm_z': 2e32,
            'power_ca': 3,
            'replenish_rate_y': 10,
            'loss_rate_l': 2580,
            'recovery_rate_r': 6580,
            'reprocess_rate_x': 66.3,
            'max_free_pool': 8,
        },
        'msr': {
            'gamma_ca': 130,
            'beta_ca': 400,
            'tau_m': 1e-4,
            'gmax_ca': 4.25e-9,
            'e_ca': 0.066,
            'tau_ca': 1e-4,
            'perm_ca0': 2.5e-11,
            'perm_z': 2e32,
            'power_ca': 3,
            'replenish_rate_y': 10,
            'loss_rate_l': 2580,
            'recovery_rate_r': 6580,
            'reprocess_rate_x': 66.3,
            'max_free_pool': 9,
        },
        'lsr': {
            'gamma_ca': 130,
            'beta_ca': 400,
            'tau_m': 1e-4,
            'gmax_ca': 2.75e-9,
            'e_ca': 0.066,
            'tau_ca': 1e-4,
            'perm_ca0': 4.2e-11,
            'perm_z': 2e32,
            'power_ca': 3,
            'replenish_rate_y': 10,
            'loss_rate_l': 2580,
            'recovery_rate_r': 6580,
            'reprocess_rate_x': 66.3,
            'max_free_pool': 6,
        }
    }

    psps = {}
    trains = []
    for cf,anf_type in itertools.product(ihcrp.keys(),anf_types):

        if (syn_mode == 'quantal') or ((cf,anf_type) not in psps):
            ### IHC and Synapse
            psp = tw.run_ihc_meddis2000(
                ihcrp=ihcrp[cf],
                fs=fs,
                syn_mode=syn_mode,
                **ihc_meddis2000_pars[anf_type]
            )
            psps[cf,anf_type] = psp

        elif (syn_mode == 'probability') and ((cf,anf_type) in psps):
            psp = psps[cf,anf_type]

        else:
            raise RuntimeError


        ### Spike generator (pars from Sumner et al. 2002)
        spikes = tw.run_an_sg_carney_holmberg2007(
            psp=psp,
            fs=fs,
            c0=0.5,
            c1=0.,
            s0=0.8e-3,
            s1=12.5e-3,
            refractory_period=0.75e-3,
        )

        trains.append({
            'spikes': spikes,
            'duration': duration,
            'cf': cf,
            'type': anf_type
        })


    spike_trains = pd.DataFrame(trains)
    return spike_trains



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo calculates modulation gain of an innear ear model.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea
from cochlea.stats import calc_modulation_gain


def main():

    gains = calc_modulation_gain(
        model=cochlea.run_zilany2014,
        model_pars={'species': 'cat'}
    )


    # gains = calc_modulation_gain(
    #     model=cochlea.run_holmberg2007,
    #     cf=cochlea.get_nearest_cf_holmberg2007(10e3),
    #     model_pars={'fs': 48e3}
    # )


    # from cochlea.external import run_matlab_auditory_periphery
    # gains = calc_modulation_gain(
    #     model=run_matlab_auditory_periphery,
    #     model_pars={'fs': 48e3}
    # )

    print(gains)

    gains.plot(logx=True)
    plt.show()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize

import numpy



with open('README.rst') as file:
    long_description = file.read()




extensions = [
    Extension(
        "cochlea.zilany2009._pycat",
        [
            "cochlea/zilany2009/_pycat.pyx",
            "cochlea/zilany2009/catmodel.c",
            "cochlea/zilany2009/complex.c"
        ]
    ),
    Extension(
        "cochlea.holmberg2007._traveling_waves",
        [
            "cochlea/holmberg2007/_traveling_waves.pyx",
        ]
    ),
    Extension(
        "cochlea.zilany2014._zilany2014",
        [
            "cochlea/zilany2014/_zilany2014.pyx",
            "cochlea/zilany2014/model_IHC.c",
            "cochlea/zilany2014/model_Synapse.c",
            "cochlea/zilany2014/complex.c"
        ]
    ),
]



setup(
    name = "cochlea",
    version = "1.4",
    author = "Marek Rudnicki",
    author_email = "marek.rudnicki@tum.de",

    description = "Inner ear models in Python",
    license = "GPLv3+",
    url = "https://github.com/mrkrd/cochlea",
    download_url = "https://github.com/mrkrd/cochlea/tarball/master",

    packages = find_packages(),
    scripts = ["scripts/run_zilany2014"],
    package_data = {
        "cochlea.asr": ["*.csv"]
    },
    include_dirs = [numpy.get_include()],
    ext_modules = cythonize(extensions),
    long_description = long_description,
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Cython",
        "Programming Language :: C",
    ],

    platforms = ["Linux", "Windows", "FreeBSD", "OSX"],
    install_requires=["numpy", "pandas", "scipy"],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This demo shows how to calculate rate-level characteristic of a
model.

"""
from __future__ import division, absolute_import, print_function

import numpy as np
import matplotlib.pyplot as plt

import cochlea
from cochlea.stats import calc_rate_level

# import cochlea.external

def main():

    # rates = calc_rate_level(
    #     model=cochlea.run_holmberg2007,
    #     cf=cochlea.get_nearest_cf_holmberg2007(1000),
    #     model_pars={'fs': 48e3, 'syn_mode': 'quantal'},
    # )

    rates = calc_rate_level(
        model=cochlea.run_zilany2014,
        cf=1000,
        model_pars={'fs': 100e3, 'species': 'human'}
    )


    # rates = calc_rate_level(
    #     model=cochlea.external.run_matlab_auditory_periphery,
    #     cf=1000,
    #     model_pars={'fs': 48e3}
    # )

    print(rates)

    rates.plot()

    plt.legend()
    plt.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import division, print_function, absolute_import

import numpy as np
import pandas as pd

from . threshold_rate import calc_thresholds_rate
from . rate_level import calc_rate_level
from . synchronization import calc_synchronization
from . modulation_gain import calc_modulation_gain
from . tuning import calc_tuning


def calc_human_hearing_thresholds(freqs):
    """Calculate human hearing thresholds for given frequencies.

    References
    ----------
    E. Terhardt, "Calculating virtual pitch", Hearing Res., vol. 1,
    pp. 155--182, 1979.

    http://www.diracdelta.co.uk/science/source/t/h/threshold%20of%20hearing/source.html

    """
    freqs = np.atleast_1d(freqs)
    f = freqs/1000

    thresholds = (
        3.64 * f**(-0.8)
        - 6.5 * np.exp(-0.6 * (f - 3.3)**2)
        + 10**(-3) * f**4
    )

    thresholds = pd.Series(thresholds, index=freqs)

    return thresholds
